#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test :any:`Object`."""

import re

from pytest import fixture, raises

import ucdp as u


@fixture
def namespace() -> u.Namespace:
    """Example Namespace."""
    return u.Namespace(
        [
            u.Param(u.IntegerType(default=8), "param"),
            u.Param(u.IntegerType(default=1), "param1"),
            u.Signal(u.UintType(16, default=15), "uint_s"),
            u.Ident(u.UintType(10), "ident0"),
            u.Ident(u.UintType(10), "ident1"),
        ]
    )


class MyExprResolver(u.ExprResolver):
    """Resolver with Array Implementation."""

    def _get_array_value(self, itemvalue: str, slice_: u.Slice) -> str:
        width = self._resolve(slice_.width, brackets=True)
        return f"{width}#{itemvalue}"


@fixture
def rslvr(namespace) -> u.ExprResolver:
    """Expression Resolver."""
    return MyExprResolver(namespace=namespace)


class MyEnumType(u.AEnumType):
    """Example Enumeration."""

    keytype: u.UintType = u.UintType(4)

    def _build(self):
        self._add(0, "a")


class MyStructType(u.AStructType):
    """Example Struct."""

    def _build(self):
        self._add("a", u.UintType(4))


def test_op(rslvr):
    """Bool Op."""
    a = u.const("2")
    b = u.const("3")

    assert rslvr.resolve(a + b) == "2 + 3"
    assert rslvr.resolve(a == b) == "2 == 3"
    assert rslvr.resolve(-a) == "-2"
    assert rslvr.resolve(abs(-a)) == "abs(-2)"


def test_slice(rslvr):
    """Resolver."""
    param = rslvr.namespace["param"]
    param1 = rslvr.namespace["param1"]
    signal = rslvr.namespace["uint_s"]

    assert rslvr.resolve(signal[2]) == "uint_s[2]"
    assert rslvr.resolve(signal[2:1]) == "uint_s[2:1]"
    assert rslvr.resolve(signal[2:0]) == "uint_s[2:0]"

    assert rslvr.resolve(signal[param]) == "uint_s[param]"
    assert rslvr.resolve(signal[param:1]) == "uint_s[param:1]"
    assert rslvr.resolve(signal[param:0]) == "uint_s[param:0]"
    assert rslvr.resolve(signal[14:param]) == "uint_s[14:param]"

    assert rslvr.resolve_slice(u.Slice(left=param)) == "[param]"

    assert rslvr.resolve(signal[param1:1]) == "uint_s[param1:1]"


def test_values(rslvr):
    """Values."""
    param = u.Param(u.BitType(), "param")
    assert rslvr.resolve(u.ConstExpr(u.BitType())) == "0"
    assert rslvr.resolve(u.ConstExpr(u.BitType(default=1))) == "1"
    assert rslvr.resolve(u.ConstExpr(u.BitType(default=param))) == "param"

    param = u.Param(u.RailType(), "param")
    assert rslvr.resolve(u.ConstExpr(u.RailType())) == ""
    assert rslvr.resolve(u.ConstExpr(u.RailType(default=0))) == "0"
    assert rslvr.resolve(u.ConstExpr(u.RailType(default=1))) == "1"
    assert rslvr.resolve(u.ConstExpr(u.RailType(default=param))) == "param"

    param = u.Param(u.BoolType(), "param")
    assert rslvr.resolve(u.ConstExpr(u.BoolType())) == "False"
    assert rslvr.resolve(u.ConstExpr(u.BoolType(default=True))) == "True"
    assert rslvr.resolve(u.ConstExpr(u.BoolType(default=param))) == "param"

    param = u.Param(u.UintType(8), "param")
    assert rslvr.resolve(u.ConstExpr(u.UintType(8))) == "0x0"
    assert rslvr.resolve(u.ConstExpr(u.UintType(8, default=1))) == "0x1"
    assert rslvr.resolve(u.ConstExpr(u.UintType(8, default=param))) == "param"

    param = u.Param(u.UintType(8), "param")
    assert rslvr.resolve(u.ConstExpr(u.UintType(8))) == "0x0"
    assert rslvr.resolve(u.ConstExpr(u.UintType(8, default=1))) == "0x1"
    assert rslvr.resolve(u.ConstExpr(u.UintType(8, default=param))) == "param"

    with raises(ValueError):
        rslvr.resolve(u.ConstExpr(u.UintType(-8)))

    param = u.Param(u.SintType(8), "param")
    assert rslvr.resolve(u.ConstExpr(u.SintType(8))) == "0x0"
    assert rslvr.resolve(u.ConstExpr(u.SintType(8, default=1))) == "0x1"
    assert rslvr.resolve(u.ConstExpr(u.SintType(8, default=-2))) == "-0x2"
    assert rslvr.resolve(u.ConstExpr(u.SintType(8, default=param))) == "param"

    param = u.Param(u.IntegerType(), "param")
    assert rslvr.resolve(u.ConstExpr(u.IntegerType())) == "0"
    assert rslvr.resolve(u.ConstExpr(u.IntegerType(default=1))) == "1"
    assert rslvr.resolve(u.ConstExpr(u.IntegerType(default=-2))) == "-2"
    assert rslvr.resolve(u.ConstExpr(u.IntegerType(default=param))) == "param"

    with raises(ValueError):
        rslvr.resolve(u.ConstExpr(u.SintType(-8)))

    param = u.Param(MyEnumType(), "param")
    assert rslvr.resolve(u.ConstExpr(MyEnumType())) == "0x0"
    assert rslvr.resolve(u.ConstExpr(MyEnumType(default=1))) == "0x1"
    assert rslvr.resolve(u.ConstExpr(MyEnumType(default=param))) == "param"

    assert rslvr.resolve(u.ConstExpr(u.StringType())) == "''"
    assert rslvr.resolve(u.ConstExpr(u.StringType(default="foo"))) == "'foo'"

    assert rslvr.resolve(u.ConstExpr(u.FloatType())) == "0.0"
    assert rslvr.resolve(u.ConstExpr(u.FloatType(default=1.2))) == "1.2"

    assert rslvr.resolve(u.ConstExpr(u.DoubleType())) == "0.0"
    assert rslvr.resolve(u.ConstExpr(u.DoubleType(default=1.2))) == "1.2"

    assert rslvr.resolve(u.TODO) == "'TODO'"

    with raises(ValueError):
        rslvr.resolve(u.ConstExpr(MyStructType()))


def test_resolve_value(rslvr):
    """Resolver."""
    assert rslvr.resolve_value(u.StringType()) == "''"
    assert rslvr.resolve_value(u.StringType(default="foo")) == "'foo'"


def test_missing(rslvr):
    """Missing Name."""
    msg = "'unknown' is not a valid expression."
    with raises(ValueError, match=re.escape(msg)):
        rslvr.resolve("unknown")

    msg = "Ident(UintType(6), 'ident') not known within current namespace."
    with raises(ValueError, match=re.escape(msg)):
        rslvr.resolve(u.Ident(u.UintType(6), "ident"))


def test_concat(rslvr):
    """Concat."""
    expr = u.ConcatExpr(
        (
            u.ConstExpr(u.UintType(5, default=5)),
            u.ConstExpr(u.UintType(10, default=0x123))[4:1],
            u.ConstExpr(u.UintType(16, default=3)),
        )
    )
    assert rslvr.resolve(expr) == "{0x5, 0x1, 0x3}"


def test_ternary(rslvr):
    """Concat."""
    param = rslvr.namespace["param"]
    ident0 = rslvr.namespace["ident0"]
    ident1 = rslvr.namespace["ident1"]
    expr = u.TernaryExpr(param == 5, ident0, ident1)
    assert rslvr.resolve(expr) == "(param == 5) ? ident0 : ident1"


def test_remap():
    """Remap."""
    param = u.Param(u.IntegerType(default=8), "param")
    width = u.Param(u.IntegerType(default=u.Log2Expr(param)), "width")
    namespace = u.Namespace(
        [
            param,
            width,
            u.Port(u.UintType(param), "data_i"),
            u.Port(u.UintType(param), "cnt_o"),
        ]
    )

    rslvr = u.ExprResolver(namespace=namespace)
    assert rslvr(param) == "param"

    remap = u.Idents([param])
    rslvr = u.ExprResolver(namespace=namespace, remap=remap)
    assert rslvr(param) == "8"
    assert rslvr(width) == "width"


def test_log2(rslvr):
    """Log2."""
    expr = u.Log2Expr(u.ConstExpr(u.UintType(8, default=8)))
    with raises(NotImplementedError):
        rslvr(expr)


def test_min(rslvr):
    """Minimum."""
    expr = u.MinimumExpr(
        (
            u.ConstExpr(u.UintType(5, default=5)),
            u.ConstExpr(u.UintType(16, default=3)),
        )
    )
    with raises(NotImplementedError):
        rslvr(expr)


def test_max(rslvr):
    """Maximum."""
    expr = u.MaximumExpr(
        (
            u.ConstExpr(u.UintType(5, default=5)),
            u.ConstExpr(u.UintType(16, default=3)),
        )
    )
    with raises(NotImplementedError):
        rslvr(expr)


def test_range(rslvr):
    """Range."""
    expr = u.RangeExpr(type_=u.UintType(4), range_=range(2, 9))
    with raises(NotImplementedError):
        rslvr(expr)


def test_array(rslvr, namespace):
    """Array."""
    param = namespace["param"]
    expr = u.ConstExpr(u.ArrayType(u.UintType(8, default=4), param))
    assert rslvr(expr) == "param#0x4"

    expr = u.ConstExpr(u.ArrayType(u.UintType(8, default=2), param * 2))
    assert rslvr(expr) == "(param * 2)#0x2"

    expr = u.ConstExpr(u.ArrayType(u.UintType(8, default=2), param * 2, left=3))
    assert rslvr(expr) == "(param * 2)#0x2"



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Doc Utilities.
"""

from .doc import Doc
from .typebase import BaseType


def doc_from_type(type_: BaseType, title: str | None = None, descr: str | None = None, comment: str | None = None):
    """
    Create [Doc][ucdp.doc.Doc] with defaults from `type_`.

    Args:
        type_: Type to derive from.

    Keyword Args:
        title: Full Spoken Name.
        descr: Documentation Description.
        comment: Source Code Comment. Default is 'title'

    ??? Example "Dictorary Examples"
        Basics:

            >>> import ucdp as u
            >>> class MyType(u.BitType):
            ...     title:str = "My Bit Title"
            ...     comment:str = "My Bit Comment"
            >>> u.doc_from_type(MyType())
            Doc(title='My Bit Title', comment='My Bit Comment')
            >>> u.doc_from_type(MyType(), title="My Title")
            Doc(title='My Title', comment='My Bit Comment')
            >>> u.doc_from_type(MyType(), title="My Title", comment="")
            Doc(title='My Title', comment='')
            >>> u.doc_from_type(MyType(), comment="My Comment")
            Doc(title='My Bit Title', comment='My Comment')
    """
    # Some kind of optimized default routine
    if title is None:
        title = type_.title
    if descr is None:
        descr = type_.descr
    if comment is None:
        comment = type_.comment
    return Doc(title=title, descr=descr, comment=comment)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Module Reference Information."""

from inspect import getfile
from pathlib import Path

from .consts import PKG_PATHS
from .mod import AMod
from .modbase import BaseMod, ModCls, ModTags
from .modconfigurable import AConfigurableMod
from .modconfigurabletb import AConfigurableTbMod
from .modcore import ACoreMod
from .modgenerictb import AGenericTbMod
from .modtailored import ATailoredMod
from .modtb import ATbMod
from .modtopref import TbType, TopModRef, get_tb, is_top
from .object import Object

BASECLSS = (AConfigurableMod, ACoreMod, ATailoredMod, AMod, AGenericTbMod, ATbMod, BaseMod, AConfigurableTbMod)


class TopModRefInfo(Object):
    """Module Reference Information."""

    topmodref: TopModRef
    tags: ModTags
    modbasecls: ModCls
    filepath: Path
    is_top: bool
    is_local: bool
    tb: TbType

    @staticmethod
    def create(topmodref: TopModRef, modcls: ModCls, tbmodcls: ModCls | None = None) -> "TopModRefInfo":
        """Create."""
        topmodcls = tbmodcls or modcls
        modfilepath = Path(getfile(modcls))
        return TopModRefInfo(
            topmodref=topmodref,
            tags=topmodcls.tags,
            modbasecls=get_modbasecls(topmodcls),
            filepath=modfilepath,
            is_top=is_top(topmodcls) or bool(tbmodcls),
            is_local=is_local(modfilepath),
            tb=get_tb(topmodcls),
        )


def get_modbasecls(modcls: ModCls) -> type[BaseMod] | None:
    """Determine Module Base Class."""
    for basecls in BASECLSS:
        if issubclass(modcls, basecls):
            return basecls
    return None  # pragma: no cover


def is_local(filepath: Path) -> bool:
    """Determine If Module Is From Local Environment And Not Installed Through A Package."""
    return not any(filepath.is_relative_to(pkg_path) for pkg_path in PKG_PATHS)



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Unified Chip Design Platform."""

from typing import ClassVar

from pydantic import ValidationError

from . import cli
from .assigns import Assign, Assigns
from .baseclassinfo import BaseClassInfo, get_baseclassinfos
from .buildproduct import ABuildProduct
from .cache import CACHE
from .casting import Casting
from .clkrel import ASYNC, ClkRel
from .clkrelbase import BaseClkRel
from .config import AConfig, AVersionConfig, BaseConfig
from .const import Const
from .consts import AUTO, PAT_IDENTIFIER, PAT_IFDEF, PAT_OPT_IDENTIFIER, PATH, RE_IDENTIFIER, RE_IFDEF, Gen
from .define import Define, Defines
from .dict import Dict
from .doc import Doc
from .docutil import doc_from_type
from .drivers import Drivers, Source, Target
from .exceptions import BuildError, DirectionError, DuplicateError, InvalidExpr, LockError, MultipleDriverError
from .expr import (
    BoolOp,
    ConcatExpr,
    ConstExpr,
    Expr,
    Log2Expr,
    MaximumExpr,
    MinimumExpr,
    Op,
    RangeExpr,
    SliceOp,
    SOp,
    TernaryExpr,
)
from .exprparser import ExprParser, cast_booltype, concat, const, log2, maximum, minimum, ternary
from .exprresolver import ExprResolver
from .filelistparser import FileListParser
from .fileset import FileSet, LibPath
from .finder import find
from .flipflop import FlipFlop
from .generate import clean, generate, get_makolator, render_generate, render_inplace
from .humannum import Bin, Bytes, Bytesize, Hex
from .ident import Ident, IdentFilter, Idents, IdentStop, get_expridents, get_ident
from .ifdef import Ifdefs, cast_ifdefs, join_ifdefs, resolve_ifdefs
from .iterutil import Names, namefilter, split
from .loader import load
from .mod import AMod
from .modbase import BaseMod, ModCls, ModClss, ModTags
from .modbasetop import BaseTopMod
from .modconfigurable import AConfigurableMod
from .modconfigurabletb import AConfigurableTbMod
from .modcore import ACoreMod
from .modfilelist import (
    Flavor,
    Flavors,
    ModFileList,
    ModFileLists,
    Paths,
    Placeholder,
    ToPaths,
    iter_modfilelists,
    resolve_modfilelist,
    resolve_modfilelists,
    search_modfilelists,
)
from .modgenerictb import AGenericTbMod
from .moditer import ModPostIter, ModPreIter, uniquemods
from .modref import ModRef
from .modtailored import ATailoredMod
from .modtb import ATbMod
from .modtopref import TopModRef
from .modtoprefinfo import TopModRefInfo
from .modutil import get_modbaseinfos, is_tb_from_modname
from .mux import Mux
from .namespace import Namespace
from .nameutil import didyoumean, get_snakecasename, join_names, split_prefix, split_suffix, str2identifier
from .note import DEFAULT, OPEN, TODO, UNUSED, Default, Note, note
from .object import (
    Field,
    IdentLightObject,
    IdentObject,
    Light,
    LightObject,
    NamedLightObject,
    NamedObject,
    Object,
    PosArgs,
    PrivateField,
    get_repr,
)
from .orientation import (
    BWD,
    FWD,
    IN,
    INOUT,
    OUT,
    AOrientation,
    Direction,
    Orientation,
)
from .overview import get_overview_tree
from .param import Param
from .pathutil import improved_glob, improved_resolve, startswith_envvar, use_envvars
from .routepath import Routeable, Routeables, RoutePath, parse_routepath, parse_routepaths
from .signal import BaseSignal, Port, Signal
from .slices import DOWN, UP, Slice, SliceDirection, mask_to_slices
from .test import Test
from .top import Top
from .typearray import ArrayType
from .typebase import ACompositeType, AScalarType, AVecType, BaseScalarType, BaseType
from .typebaseenum import BaseEnumType, EnumItem, EnumItemFilter
from .typeclkrst import ClkRstAnType, ClkType, DiffClkRstAnType, DiffClkType, RstAnType, RstAType, RstType
from .typedescriptivestruct import DescriptiveStructType
from .typeenum import AEnumType, AGlobalEnumType, BusyType, DisType, DynamicEnumType, EnaType
from .typefloat import DoubleType, FloatType
from .typescalar import BitType, BoolType, IntegerType, RailType, SintType, UintType
from .typestring import StringType
from .typestruct import (
    AGlobalStructType,
    AStructType,
    BaseStructType,
    DynamicStructType,
    StructFilter,
    StructItem,
    bwdfilter,
    fwdfilter,
)
from .typeutil import is_scalar, is_signed
from .util import extend_sys_path, get_copyright

__all__ = [
    "ASYNC",
    "AUTO",
    "BWD",
    "CACHE",
    "DEFAULT",
    "DOWN",
    "FWD",
    "IN",
    "INOUT",
    "OPEN",
    "OUT",
    "PATH",
    "PAT_IDENTIFIER",
    "PAT_IFDEF",
    "PAT_OPT_IDENTIFIER",
    "RE_IDENTIFIER",
    "RE_IFDEF",
    "TODO",
    "UNUSED",
    "UP",
    "ABuildProduct",
    "ACompositeType",
    "AConfig",
    "AConfigurableMod",
    "AConfigurableTbMod",
    "ACoreMod",
    "AEnumType",
    "AGenericTbMod",
    "AGlobalEnumType",
    "AGlobalStructType",
    "AMod",
    "AOrientation",
    "AScalarType",
    "AStructType",
    "ATailoredMod",
    "ATbMod",
    "AVecType",
    "AVersionConfig",
    "ArrayType",
    "Assign",
    "Assigns",
    "BaseClassInfo",
    "BaseClkRel",
    "BaseConfig",
    "BaseEnumType",
    "BaseMod",
    "BaseScalarType",
    "BaseSignal",
    "BaseStructType",
    "BaseTopMod",
    "BaseType",
    "Bin",
    "BitType",
    "BoolOp",
    "BoolType",
    "BuildError",
    "BusyType",
    "Bytes",
    "Bytesize",
    "Casting",
    "ClassVar",
    "ClkRel",
    "ClkRstAnType",
    "ClkType",
    "ConcatExpr",
    "Const",
    "ConstExpr",
    "Default",
    "Define",
    "Define",
    "Defines",
    "DescriptiveStructType",
    "Dict",
    "DiffClkRstAnType",
    "DiffClkType",
    "Direction",
    "DirectionError",
    "DisType",
    "Doc",
    "DoubleType",
    "Drivers",
    "DuplicateError",
    "DynamicEnumType",
    "DynamicStructType",
    "EnaType",
    "EnumItem",
    "EnumItemFilter",
    "Expr",
    "ExprParser",
    "ExprResolver",
    "Field",
    "FileListParser",
    "FileSet",
    "Flavor",
    "Flavors",
    "FlipFlop",
    "FloatType",
    "Gen",
    "Hex",
    "Ident",
    "IdentFilter",
    "IdentLightObject",
    "IdentObject",
    "IdentStop",
    "Idents",
    "Ifdefs",
    "IntegerType",
    "InvalidExpr",
    "LibPath",
    "Light",
    "LightObject",
    "LockError",
    "Log2Expr",
    "MaximumExpr",
    "MinimumExpr",
    "ModCls",
    "ModClss",
    "ModFileList",
    "ModFileLists",
    "ModPostIter",
    "ModPreIter",
    "ModRef",
    "ModTags",
    "MultipleDriverError",
    "Mux",
    "NamedLightObject",
    "NamedObject",
    "Names",
    "Namespace",
    "Note",
    "Object",
    "Op",
    "Orientation",
    "Param",
    "Paths",
    "Placeholder",
    "Port",
    "PosArgs",
    "PrivateField",
    "RailType",
    "RangeExpr",
    "RoutePath",
    "Routeable",
    "Routeables",
    "RstAType",
    "RstAnType",
    "RstType",
    "SOp",
    "Signal",
    "SintType",
    "Slice",
    "SliceDirection",
    "SliceOp",
    "Source",
    "StringType",
    "StructFilter",
    "StructItem",
    "Target",
    "TernaryExpr",
    "Test",
    "ToPaths",
    "Top",
    "TopModRef",
    "TopModRefInfo",
    "UintType",
    "ValidationError",
    "bwdfilter",
    "cast_booltype",
    "cast_ifdefs",
    "clean",
    "cli",
    "concat",
    "const",
    "didyoumean",
    "doc_from_type",
    "extend_sys_path",
    "find",
    "fwdfilter",
    "generate",
    "get_baseclassinfos",
    "get_copyright",
    "get_expridents",
    "get_ident",
    "get_makolator",
    "get_modbaseinfos",
    "get_overview_tree",
    "get_repr",
    "get_snakecasename",
    "improved_glob",
    "improved_resolve",
    "is_scalar",
    "is_signed",
    "is_tb_from_modname",
    "iter_modfilelists",
    "join_ifdefs",
    "join_names",
    "load",
    "log2",
    "mask_to_slices",
    "maximum",
    "minimum",
    "namefilter",
    "note",
    "parse_routepath",
    "parse_routepaths",
    "render_generate",
    "render_inplace",
    "resolve_ifdefs",
    "resolve_modfilelist",
    "resolve_modfilelists",
    "search_modfilelists",
    "split",
    "split_prefix",
    "split_suffix",
    "startswith_envvar",
    "str2identifier",
    "ternary",
    "uniquemods",
    "use_envvars",
]



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Module File Information."""

from pathlib import Path

import ucdp as u

TESTS_PATH = Path(__file__).parent


def test_basic(example_simple):
    """Basic Testing."""
    from fileliststandard import HdlFileList
    from uart_lib.uart import UartMod

    mod = UartMod()
    prjroot = Path("$PRJROOT")
    filepath = prjroot / "uart_lib" / "uart" / "rtl" / "uart.sv"
    assert u.resolve_modfilelist(mod, "hdl") == HdlFileList(
        gen="full",
        filepaths=(filepath,),
        template_filepaths=(example_simple / "uart_lib" / "main.mako",),
        clean_filepaths=(),
    )


# def test_target(example_simple):
#     """Basic Testing."""
#     from glbl.clk_gate import ClkGateMod

#     mod = ClkGateMod()

#     filepath = example_simple / "glbl" / "clk_gate.sv"
#     modfilelist = u.resolve_modfilelist(mod, "hdl")
#     assert modfilelist == u.ModFileList(name="hdl", gen="inplace", target="!asic", filepaths=(filepath,))

#     filepath = example_simple / "glbl" / "clk_gate_asic.sv"
#     modfilelist = u.resolve_modfilelist(mod, "hdl", target="asic")
#     assert modfilelist == u.ModFileList(name="hdl", gen="inplace", target="asic", filepaths=(filepath,))

#     filepath = example_simple / "glbl" / "clk_gate.sv"
#     modfilelist = u.resolve_modfilelist(mod, "hdl", target="fpga")
#     assert modfilelist == u.ModFileList(name="hdl", gen="inplace", target="!asic", filepaths=(filepath,))


def test_filelistparser(example_filelist):
    """File List."""
    from filelist_lib.filelist import FilelistMod

    mod = FilelistMod()
    assert u.resolve_modfilelist(mod, "hdl") == u.ModFileList(
        name="hdl",
        filepaths=(
            example_filelist / "filelist_lib" / "mod0.sv",
            example_filelist / "filelist_lib" / "sub" / "mod2.sv",
            example_filelist / "mod2.sv",
        ),
        inc_dirs=(example_filelist / "filelist_lib" / "inc",),
    )


class MultiMod(u.AMod):
    """Just An Example Module."""

    filelists: u.ClassVar[u.ModFileLists] = (
        u.ModFileList(name="drvhpp", gen="full", filepaths=("drv.hpp",), template_filepaths=("drv.hpp.mako",)),
        u.ModFileList(name="drvcpp", gen="full", filepaths=("drv.cpp",), template_filepaths=("drv.cpp.mako",)),
    )

    def _build(self) -> None:
        pass


def test_multi():
    """Multi Module."""
    mod = MultiMod()
    assert tuple(u.resolve_modfilelists(mod, "drv*")) == (
        u.ModFileList(
            name="drvhpp",
            gen="full",
            filepaths=(TESTS_PATH / "drv.hpp",),
            template_filepaths=(TESTS_PATH / "drv.hpp.mako",),
        ),
        u.ModFileList(
            name="drvcpp",
            gen="full",
            filepaths=(TESTS_PATH / "drv.cpp",),
            template_filepaths=(TESTS_PATH / "drv.cpp.mako",),
        ),
    )


class FlavorMod(u.AMod):
    """Just An Example Module."""

    filelists: u.ClassVar[u.ModFileLists] = (
        u.ModFileList(
            name="hdl",
            gen="full",
            filepaths=("{mod.modname}_{flavor}.sv",),
            template_filepaths=("{flavor}.sv",),
            flavors=("one", "two"),
        ),
    )

    def _build(self) -> None:
        pass


def test_flavor():
    """Flavor Module."""
    mod = FlavorMod()
    assert tuple(u.resolve_modfilelists(mod, "*")) == (
        u.ModFileList(
            name="hdl",
            gen="full",
            filepaths=(TESTS_PATH / "flavor_one.sv",),
            template_filepaths=(TESTS_PATH / "one.sv",),
            flavors=("one",),
            flavor="one",
        ),
        u.ModFileList(
            name="hdl",
            gen="full",
            filepaths=(TESTS_PATH / "flavor_two.sv",),
            template_filepaths=(TESTS_PATH / "two.sv",),
            flavors=("two",),
            flavor="two",
        ),
    )



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Top."""

from .iterutil import Names
from .modbase import BaseMod
from .modgenerictb import AGenericTbMod
from .moditer import FilterFunc, MaxLevel, ModPostIter, ModPreIter, StopFunc, get_mod, get_mods
from .modtopref import TopModRef
from .object import _CACHED_INSTANCES, Field, Object


class Top(Object):
    """
    Top Module Reference.

    Attributes:
        ref: Top Module Reference
        mod: Top Module Instance
    """

    ref: TopModRef = Field(repr=False)
    mod: BaseMod

    @staticmethod
    def from_mod(mod: BaseMod) -> "Top":
        """Create from Module."""
        if isinstance(mod, AGenericTbMod):
            ref = TopModRef(tb=mod.get_modref(), top=mod.dut.get_modref())
        else:
            ref = TopModRef(top=mod.get_modref())
        return Top(ref=ref, mod=mod)

    def iter(
        self,
        filter_: FilterFunc | None = None,
        stop: StopFunc | None = None,
        maxlevel: MaxLevel | None = None,
        unique: bool = False,
        post: bool = False,
    ):
        """
        Iterate Over All Modules.

        Args:
            filter_: function called with every `mod` as argument, `mod` is returned if `True`.
            stop: stop iteration at `mod` if `stop` function returns `True` for `mod`.
            maxlevel (int): maximum descending in the mod hierarchy.
            unique (bool): Just return module once and **NOT** all instances of it.
            post: Post-Order Iteration Strategy instead of Pre-Order Iteration Strategy.
        """
        if post:
            return ModPostIter(self.mod, filter_=filter_, stop=stop, maxlevel=maxlevel, unique=unique)
        return ModPreIter(self.mod, filter_=filter_, stop=stop, maxlevel=maxlevel, unique=unique)

    def get_mods(self, namepats: Names | None = None, unique: bool = False, base: bool = False):
        """
        Return all modules matching `namepats`.

        Iterate top and all its submodules and return matching ones.

        Keyword Args:
            namepats: Iterable with name pattern (including `*` and `?`) or comma separated string
            unique (bool): Just return every module once.
            base: namepats must match against module `basequalnames` instead of `qual_name`.
        """
        return get_mods(self.mod, namepats=namepats, unique=unique, base=base)

    def get_mod(self, namepats: Names, base: bool = False):
        """
        Return the one and just the one hardware module matching `namepats`.

        Iterate over `mod` and all its submodules and return matching one.

        Keyword Args:
            namepats: Iterable with name pattern (including `*` and `?`) or comma separated string
            base: namepats must match against module `basequalnames` instead of `qual_name`.
        """
        return get_mod(self.mod, namepats, base=base)

    def get_stat(self) -> dict[str, int]:
        """Get Statistics."""
        return {
            "Modules": len(self.get_mods(unique=True)),
            "Module-Instances": len(self.get_mods()),
            "LightObjects": _CACHED_INSTANCES,
        }



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Expression Testing."""

from test2ref import assert_refdata

import ucdp as u


def _test_op(file, expr):
    type_ = getattr(expr, "type_", "-")
    file.write(f"{expr!r} {int(expr)} {bool(expr)} {type_}\n")


def test_const_const(tmp_path):
    """Const op Const."""
    one = u.const("16'd10")
    other = u.const("16'd5")
    resultfile = tmp_path / "result.txt"
    with resultfile.open("w") as file:
        _test_op(file, one + other)
        _test_op(file, one - other)
        _test_op(file, one * other)
        _test_op(file, one // other)
        _test_op(file, one / other)
        _test_op(file, one % other)
        _test_op(file, one**other)
        _test_op(file, one > other)
        _test_op(file, one >= other)
        _test_op(file, one >> other)
        _test_op(file, one < other)
        _test_op(file, one <= other)
        _test_op(file, one << other)
        _test_op(file, one == other)
        _test_op(file, one != other)
        _test_op(file, one & other)
        _test_op(file, one | other)
        _test_op(file, one ^ other)
        _test_op(file, ~one)
        _test_op(file, -one)
        _test_op(file, abs(one))
        _test_op(file, abs(-one))
    assert_refdata(test_const_const, tmp_path)


def test_const_int(tmp_path):
    """Const op int."""
    one = u.const("16'd10")
    other = 5
    resultfile = tmp_path / "result.txt"
    with resultfile.open("w") as file:
        _test_op(file, one + other)
        _test_op(file, one - other)
        _test_op(file, one * other)
        _test_op(file, one // other)
        _test_op(file, one / other)
        _test_op(file, one % other)
        _test_op(file, one**other)
        _test_op(file, one > other)
        _test_op(file, one >= other)
        _test_op(file, one >> other)
        _test_op(file, one < other)
        _test_op(file, one <= other)
        _test_op(file, one << other)
        _test_op(file, one == other)
        _test_op(file, one != other)
        _test_op(file, one & other)
        _test_op(file, one | other)
        _test_op(file, one ^ other)
        _test_op(file, ~one)
        _test_op(file, -one)
        _test_op(file, abs(one))
        _test_op(file, abs(-one))
    assert_refdata(test_const_int, tmp_path)


def test_int_const(tmp_path):
    """Const op int."""
    one = 10
    other = u.const("16'd5")
    resultfile = tmp_path / "result.txt"
    with resultfile.open("w") as file:
        _test_op(file, one + other)
        _test_op(file, one - other)
        _test_op(file, one * other)
        _test_op(file, one // other)
        _test_op(file, one / other)
        _test_op(file, one % other)
        _test_op(file, one**other)
        _test_op(file, one > other)
        _test_op(file, one >= other)
        _test_op(file, one >> other)
        _test_op(file, one < other)
        _test_op(file, one <= other)
        _test_op(file, one << other)
        _test_op(file, one == other)
        _test_op(file, one != other)
        _test_op(file, one & other)
        _test_op(file, one | other)
        _test_op(file, one ^ other)
        _test_op(file, ~one)
        _test_op(file, -one)
        _test_op(file, abs(one))
        _test_op(file, abs(-one))
    assert_refdata(test_int_const, tmp_path)


def test_slice_const():
    """Test Slicing of Const."""
    param_p = u.ConstExpr(u.UintType(8, default=5))
    expr = u.const(2)
    assert param_p[2:1] == u.SliceOp(u.Param(u.UintType(8, default=5), "param_p"), u.Slice("2:1"))
    assert param_p[2] == u.SliceOp(u.Param(u.UintType(8, default=5), "param_p"), u.Slice("2"))
    assert param_p[expr] == u.SliceOp(
        u.Param(u.UintType(8, default=5), "param_p"), u.Slice(u.ConstExpr(u.IntegerType(default=2)))
    )


def test_slice_param():
    """Test Slicing of Param."""
    param_p = u.Param(u.UintType(8, default=5), "param_p")
    expr = u.const(2)
    assert param_p[2:1] == u.SliceOp(u.Param(u.UintType(8, default=5), "param_p"), u.Slice("2:1"))
    assert param_p[2] == u.SliceOp(u.Param(u.UintType(8, default=5), "param_p"), u.Slice("2"))
    assert param_p[expr] == u.SliceOp(
        u.Param(u.UintType(8, default=5), "param_p"), u.Slice(u.ConstExpr(u.IntegerType(default=2)))
    )



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Expression Engine.
"""

import math
import operator
import re
from collections.abc import Callable, Iterator
from typing import ClassVar

from icdutil.num import calc_signed_width, calc_unsigned_width, unsigned_to_signed
from pydantic import ValidationError

from .exceptions import InvalidExpr
from .object import Field, Light, Object, PosArgs, model_validator
from .slices import Slice
from .typebase import BaseScalarType, BaseType
from .typehelper import BaseScalarTypes
from .typescalar import BitType, BoolType, IntegerType, SintType, UintType

_RE_CONST = re.compile(
    r"(?P<sign>[-+])?"
    r"(((?P<width>\d+)?'?(?P<is_signed>s)?(?P<bnum>(b[01_]+)|(o[0-7_]+)|(d[0-9_]+)|(h[0-9a-fA-F_]+))))|(?P<num>[+-]?\d+)\b"
)
_NUM_BASEMAP = {
    "b": 2,
    "o": 8,
    "d": 10,
    "h": 16,
    None: 10,
}
_WIDTH_BASEMAP = {
    "b": 1,
    "o": 3,
    "d": math.log(10) / math.log(2),
    "h": 4,
}
_OPERMAP = {
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    ">": operator.gt,
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "//": operator.floordiv,
    "/": operator.truediv,
    "%": operator.mod,
    "**": operator.pow,
    "<<": operator.lshift,
    ">>": operator.rshift,
    "|": operator.or_,
    "&": operator.and_,
    "^": operator.xor,
}
_SOPERMAP = {
    "abs(": operator.abs,
    "~": operator.inv,
    "-": operator.neg,
}


class Expr(Object):  # noqa: PLW1641
    """Base Class for all Expressions.

    Attributes:
        type_: Type.
    """

    type_: BaseType = Field(repr=False)

    def __lt__(self, other):
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return BoolOp(self, "<", other)

    def __le__(self, other):
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return BoolOp(self, "<=", other)

    def __eq__(self, other):
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return BoolOp(self, "==", other)

    def __ne__(self, other):
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return BoolOp(self, "!=", other)

    def __ge__(self, other):
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return BoolOp(self, ">=", other)

    def __gt__(self, other):
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return BoolOp(self, ">", other)

    def __add__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(self, "+", other)

    def __sub__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(self, "-", other)

    def __mul__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(self, "*", other)

    def __floordiv__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(self, "//", other)

    def __truediv__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(self, "/", other)

    def __mod__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(self, "%", other)

    def __pow__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(self, "**", other)

    def __lshift__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(self, "<<", other)

    def __rshift__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(self, ">>", other)

    def __or__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(self, "|", other)

    def __and__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(self, "&", other)

    def __xor__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(self, "^", other)

    def __radd__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(other, "+", self)

    def __rsub__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(other, "-", self)

    def __rmul__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(other, "*", self)

    def __rfloordiv__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(other, "//", self)

    def __rtruediv__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(other, "/", self)

    def __rmod__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(other, "%", self)

    def __rpow__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(other, "**", self)

    def __rlshift__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(other, "<<", self)

    def __rrshift__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(other, ">>", self)

    def __ror__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(other, "|", self)

    def __rand__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(other, "&", self)

    def __rxor__(self, other) -> "Op":
        if isinstance(other, int):
            other = _parse_const(other, reftype=self.type_)
        if not isinstance(other, Expr):
            return NotImplemented
        return Op(other, "^", self)

    def __abs__(self) -> "SOp":
        return SOp(sign="abs(", one=self)

    def __invert__(self) -> "SOp":
        return SOp(sign="~", one=self)

    def __neg__(self) -> "SOp":
        return SOp(sign="-", one=self)

    def __getitem__(self, slice_):
        if isinstance(slice_, Expr):
            slice_ = Slice(left=slice_, right=slice_)
        else:
            slice_ = Slice.cast(slice_)
        return SliceOp(one=self, slice_=slice_)


class Op(Expr, Light):
    """Dual Operator Expression.

    Args:
        left: left argument.
        sign: sign.
        right: right argument.

    Attributes:
        oper: Operator.
        type_: Type.

    ???+ bug "Todo"
        * fix inherited_members / Attributes Types
    """

    left: Expr
    oper: Callable = Field(repr=False)
    sign: str
    right: Expr

    _posargs: ClassVar[PosArgs] = ("left", "sign", "right")

    def __init__(self, left: Expr, sign: str, right: Expr, type_: BaseType | None = None):
        oper = _OPERMAP[sign]
        if type_ is None:
            type_ = left.type_
        super().__init__(left=left, oper=oper, sign=sign, right=right, type_=type_)  # type: ignore[call-arg]

    def __int__(self):
        return int(self.oper(int(self.left), int(self.right)))

    def __bool__(self):
        return bool(self.oper(int(self.left), int(self.right)))


class BoolOp(Op):
    """Boolean Dual Operator Expression.

    Args:
        left: left argument.
        sign: sign.
        right: right argument.

    Attributes:
        oper: Operator.
        type_: Type.

    ???+ bug "Todo"
        * fix inherited_members / Attributes Types
    """

    def __init__(self, left: Expr, sign: str, right: Expr):
        super().__init__(left=left, sign=sign, right=right, type_=BoolType())  # type: ignore[call-arg]


class SOp(Expr, Light):
    """Single Operator Expression.

    Args:
        sign: sign.
        one: Expression.

    Attributes:
        oper: Operator.
        postsign: postsign.
        type_: Type.

    ???+ bug "Todo"
        * fix inherited_members / Attributes Types

    """

    oper: Callable = Field(repr=False)
    sign: str
    one: Expr
    postsign: str = ""

    _posargs: ClassVar[PosArgs] = ("sign", "one")

    def __init__(self, sign: str, one: Expr):
        oper = _SOPERMAP[sign]
        postsign = ")" if "(" in sign else ""
        super().__init__(oper=oper, sign=sign, one=one, postsign=postsign, type_=one.type_)  # type: ignore[call-arg]

    def __int__(self):
        return self.oper(int(self.one))


class SliceOp(Expr, Light):
    """Slice Expression.

    Args:
        one: Expression.
        slice_: Slice

    Attributes:
        type_: Type.

    ???+ bug "Todo"
        * fix inherited_members / Attributes Types
    """

    one: Expr
    slice_: Slice

    _posargs: ClassVar[PosArgs] = ("one", "slice_")

    def __init__(self, one: Expr, slice_: Slice):
        super().__init__(one=one, slice_=slice_, type_=one.type_[slice_])  # type: ignore[call-arg]

    def __int__(self):
        return int(self.one.type_[self.slice_].default)


class ConstExpr(Expr, Light):
    """
    Constant.

    Args:
        type_: Type.

    ??? Example "ConstExpr Examples"
        Example.

            >>> import ucdp as u
            >>> const = u.ConstExpr(u.UintType(5, default=5))
            >>> const
            ConstExpr(UintType(5, default=5))
            >>> int(const)
            5
            >>> bool(const)
            True
    """

    type_: BaseType

    _posargs: ClassVar[PosArgs] = ("type_",)

    def __init__(self, type_: BaseType):
        super().__init__(type_=type_)  # type: ignore[call-arg]

    def __int__(self):
        return int(self.type_.default)

    def __getitem__(self, slice_):
        if isinstance(slice_, Expr):
            slice_ = Slice(left=slice_, right=slice_)
        return ConstExpr(self.type_[slice_])


class ConcatExpr(Expr, Light):
    """
    Concatenation.

    Args:
        items: Expressions.

    Attributes:
        type_: Type.

    ???+ bug "Todo"
        * fix inherited_members / Attributes Types

    ??? Example "ConcatExpr Examples"
        Example.

            >>> import ucdp as u
            >>> expr = u.ConcatExpr((
            ...     u.ConstExpr(u.UintType(5, default=5)),
            ...     u.ConstExpr(u.UintType(7, default=1)),
            ...     u.ConstExpr(u.UintType(16, default=3)),
            ... ))
            >>> expr
            ConcatExpr((ConstExpr(UintType(5, default=5)), ... ConstExpr(UintType(16, default=3))))
            >>> int(expr)
            12325
    """

    items: tuple[Expr, ...]

    _posargs: ClassVar[PosArgs] = ("items",)

    def __init__(self, items: tuple[Expr, ...]):
        for item in items:
            if not isinstance(item.type_, BaseScalarType):
                raise ValueError(f"Item {item} type is {item.type_}, but must be a Scalar Type")

        pairs = tuple(ConcatExpr.__iter_values(items))
        default = sum(value << shift for value, shift in pairs)
        width = pairs[-1][1] if pairs else 1
        type_ = UintType(width, default=default)
        super().__init__(items=items, type_=type_)  # type: ignore[call-arg]

    def __int__(self):
        return sum(int(value << shift) for value, shift in self.__iter_values(self.items))

    @staticmethod
    def __iter_values(items: tuple[Expr, ...]) -> Iterator[tuple[int, int]]:
        shift = 0
        for item in items:
            yield item, shift
            shift += item.type_.width  # type: ignore[attr-defined]
        yield 0, shift


class TernaryExpr(Expr, Light):
    """
    TernaryExpr Expression.

    Args:
        cond: BoolOp
        one: Expression
        other: Expression

    Attributes:
        type_: Type.

    ??? Example "TernaryExpr Examples"
        Example.

            >>> import ucdp as u
            >>> cond = u.Signal(u.UintType(2), 'if_s') == u.ConstExpr(UintType(2, default=1))
            >>> one = u.Signal(u.UintType(16, default=10), 'one_s')
            >>> other = u.Signal(u.UintType(16, default=20), 'other_s')
            >>> expr = TernaryExpr(cond=cond, one=one, other=other)
            >>> expr
            TernaryExpr(BoolOp(Signal(UintType(2), 'if_s'), ... Signal(UintType(16, default=20), 'other_s'))
            >>> int(expr)
            20
            >>> expr.type_
            UintType(16, default=10)
    """

    type_: BaseScalarTypes = Field(repr=False)
    cond: BoolOp
    one: Expr
    other: Expr

    _posargs: ClassVar[PosArgs] = ("cond", "one", "other")

    def __init__(self, cond: BoolOp, one: Expr, other: Expr):
        super().__init__(cond=cond, one=one, other=other, type_=one.type_)  # type: ignore[call-arg]

    def __int__(self):
        if bool(self.cond):
            return int(self.one)
        return int(self.other)


class Log2Expr(Expr, Light):
    """
    Ceiling Logarithm to base of 2.

    Args:
        expr: Expression

    Attributes:
        type_: Type.

    ??? Example "Log2Expr Examples"
        Example.

            >>> import ucdp as u
            >>> expr = u.Log2Expr(u.ConstExpr(u.UintType(5, default=5)))
            >>> expr
            Log2Expr(ConstExpr(UintType(5, default=5)))
            >>> int(expr)
            2
            >>> expr.type_
            UintType(5, default=5)
    """

    type_: BaseScalarTypes = Field(repr=False)
    expr: Expr

    _posargs: ClassVar[PosArgs] = ("expr",)

    def __init__(self, expr: Expr):
        super().__init__(expr=expr, type_=expr.type_)  # type: ignore[call-arg]

    def __int__(self):
        return int(math.log(int(self.expr), 2))


class MinimumExpr(Expr, Light):
    """
    Smallest Value.

    Args:
        items: Items

    Attributes:
        type_: Type.

    ??? Example "MinimumExpr Examples"
        Example.

            >>> import ucdp as u
            >>> expr = u.MinimumExpr((
            ...     u.ConstExpr(u.UintType(5, default=5)),
            ...     u.ConstExpr(u.UintType(7, default=1)),
            ...     u.ConstExpr(u.UintType(16, default=3)),
            ... ))
            >>> expr
            MinimumExpr((ConstExpr(UintType(5, default=5)), ... ConstExpr(UintType(16, default=3))))
            >>> int(expr)
            1
            >>> expr.type_
            UintType(5, default=5)
    """

    type_: BaseScalarTypes = Field(repr=False)
    items: tuple[Expr, ...]

    _posargs: ClassVar[PosArgs] = ("items",)

    def __init__(self, items: tuple[Expr, ...]):
        super().__init__(items=items, type_=items[0].type_)  # type: ignore[call-arg]

    def __int__(self):
        return min(int(item) for item in self.items)


class MaximumExpr(Expr, Light):
    """
    Largest Value.

    Args:
        items: Items

    Attributes:
        type_: Type.

    ??? Example "MaximumExpr Examples"
        Example.

            >>> import ucdp as u
            >>> expr = u.MaximumExpr((
            ...     u.ConstExpr(u.UintType(5, default=5)),
            ...     u.ConstExpr(u.UintType(7, default=1)),
            ...     u.ConstExpr(u.UintType(16, default=3)),
            ... ))
            >>> expr
            MaximumExpr((ConstExpr(UintType(5, default=5)), ... ConstExpr(UintType(16, default=3))))
            >>> int(expr)
            5
            >>> expr.type_
            UintType(5, default=5)
    """

    type_: BaseScalarTypes = Field(repr=False)
    items: tuple[Expr, ...]

    _posargs: ClassVar[PosArgs] = ("items",)

    def __init__(self, items: tuple[Expr, ...]):
        super().__init__(items=items, type_=items[0].type_)  # type: ignore[call-arg]

    def __int__(self):
        return max(int(item) for item in self.items)


class RangeExpr(Expr, Light):
    """
    Value Range.

    Attributes:
        type_: Type.
        range_: Range.

    ??? Example "RangeExpr Examples"
        Example.

            >>> import ucdp as u
            >>> range_ = u.RangeExpr(type_=u.UintType(4), range_=range(2, 9))
            >>> range_.type_
            UintType(4)
            >>> range_.range_
            range(2, 9)
    """

    type_: UintType | SintType
    range_: range

    @model_validator(mode="after")
    def __post_init(self) -> "RangeExpr":
        values = tuple(self.range_)
        self.type_.check(values[0], what="Start Value")
        self.type_.check(values[-1], what="End Value")
        return self


def _parse_const(value, reftype: BaseType | None = None) -> ConstExpr:
    strippedvalue = str(value).strip()
    matnum = _RE_CONST.fullmatch(strippedvalue)
    if matnum:
        return __parse_const(reftype=reftype, **matnum.groupdict())
    raise InvalidExpr(repr(value))


def __parse_const(sign, width, is_signed, bnum, num, reftype) -> ConstExpr:  # noqa: C901
    # Bin/Oct/Dec/Hex Number with given width
    if num is None:
        base, num = bnum[0], bnum[1:].replace("_", "")
        value = int(num, _NUM_BASEMAP[base])
        if sign == "-":
            value = -value
        if width is None:
            width = math.ceil(_WIDTH_BASEMAP[base] * len(num))
        else:
            width = int(width)
        type_: BaseType
        if base == "b" and width == 1 and not is_signed:
            type_ = BitType(default=value)
        elif is_signed:
            if value > 0:
                value = unsigned_to_signed(value, width)
            type_ = SintType(width, default=value)
        else:
            type_ = UintType(width, default=value)
        return ConstExpr(type_)

    # width-less integer
    intnum = int(num)
    if reftype is not None:
        try:
            return ConstExpr(reftype.new(default=intnum))
        except ValidationError:
            pass

    # Integer
    if IntegerType.min_ <= intnum <= IntegerType.max_:
        return ConstExpr(IntegerType(default=intnum))

    # signed vector
    if intnum < 0:
        width = calc_signed_width(intnum)
        return ConstExpr(SintType(width, default=intnum))

    # unsigned vector
    width = calc_unsigned_width(intnum)
    return ConstExpr(UintType(width, default=intnum))



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Module File Information."""

from pathlib import Path

import ucdp as u


def test_basic(example_simple):
    """Basic Testing."""
    from uart_lib.uart import UartMod

    mod = UartMod()

    info = u.FileSet.from_mod(mod, "hdl")
    prjroot = Path("$PRJROOT")
    assert info.target is None
    assert info.filepaths == (
        u.LibPath(libname="glbl_lib", path=prjroot / "glbl_lib" / "clk_gate" / "rtl" / "clk_gate.sv"),
        u.LibPath(libname="uart_lib", path=prjroot / "uart_lib" / "uart" / "rtl" / "uart_regf.sv"),
        u.LibPath(libname="uart_lib", path=prjroot / "uart_lib" / "uart" / "rtl" / "uart_core.sv"),
        u.LibPath(libname="uart_lib", path=prjroot / "uart_lib" / "uart" / "rtl" / "uart.sv"),
    )
    assert info.inc_dirs == ()

    info = u.FileSet.from_mod(mod, "systemc")
    assert info.target is None
    assert info.filepaths == ()
    assert info.inc_dirs == ()



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Loader.

* [load()][ucdp.loader.load] is one and only method to pickup and instantiate the topmost hardware module.
"""

from collections.abc import Iterable
from pathlib import Path

from ._modloader import build_top, find_modrefs, load_modcls
from .logging import LOGGER
from .modbase import BaseMod
from .modgenerictb import AGenericTbMod
from .moditer import get_mod
from .modref import ModRef
from .modtopref import TopModRef
from .modtoprefinfo import get_modbasecls, is_top
from .nameutil import didyoumean
from .top import Top
from .util import extend_sys_path


def load(topmodref: TopModRef | Path | str, paths: Iterable[Path] | None = None) -> Top:
    """
    Load Module from ``topmodref`` and return :any:`Top`.

    In case of a given ``topref.sub`` search for a submodule named ``sub`` within the
    module hierarchy of ``topmod`` using :any:`Top.get_mod()`.

    In case of a given ``tb`` search for a testbench ``tb`` and pair it.

    Args:
        topmodref: Items.

    Keyword Args:
        paths: Additional Search Paths for Python Modules. UCDP_PATHS environment variable by default.

    Returns:
        Top: Top
    """
    with extend_sys_path(paths, use_env_default=True):
        topmodref = TopModRef.cast(topmodref)
        mod = _load_topmod(topmodref)
        return Top(ref=topmodref, mod=mod)


def _load_topmod(ref: TopModRef) -> BaseMod:
    LOGGER.info("Loading %r", str(ref))

    modcls = _load_modcls_dym(ref.top)
    if not is_top(modcls):
        modbasecls = get_modbasecls(modcls)
        raise ValueError(f"{ref.top} is not a top module as it bases on {modbasecls}")
    mod = build_top(modcls)
    if ref.sub:
        mod = get_mod(mod, ref.sub)
    if ref.tb:
        tbcls = _load_modcls_dym(ref.tb)
        if not issubclass(tbcls, AGenericTbMod):
            raise ValueError(f"{tbcls} is not a testbench module aka child of <class ucdp.AGenericTbMod>.")
        return tbcls.build_tb(mod)
    return mod


def _load_modcls_dym(modref: ModRef) -> type[BaseMod]:
    """Load Module Class."""
    try:
        return load_modcls(modref)
    except NameError as exc:
        modrefs = [str(modref) for modref in find_modrefs()]
        dym = didyoumean(str(modref), modrefs)
        raise NameError(f"{exc!s}{dym}") from None



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Param."""

import ucdp as u


class ParamMod(u.AMod):
    """Example Module With Parameters."""

    def _build(self):
        parser = self.parser
        # a module parameter which can be set at instantiation
        param_p = self.add_param(u.IntegerType(default=10), "param_p")
        # a parameter which is derived by a formula from a previous parameter
        width_p = self.add_param(u.IntegerType(default=parser.log2(param_p + 1)), "width_p")
        # just another parameter
        default_p = self.add_param(u.IntegerType(default=param_p), "default_p")

        # ports which depend on the parameter above
        self.add_port(u.UintType(param_p), "data_i")
        self.add_port(u.UintType(width_p), "cnt_o")

        # an internal constant
        self.add_const(u.UintType(param_p, default=default_p), "const_c")



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test :any:`Object`."""

import re

from pytest import fixture, raises

import ucdp as u


@fixture
def idents() -> u.Idents:
    """Some Identifier."""
    return u.Idents(
        [
            u.Signal(u.UintType(16, default=15), "uint_s"),
            u.Signal(u.SintType(16, default=-15), "sint_s"),
        ]
    )


@fixture
def parser(idents) -> u.ExprParser:
    """Some Identifier."""
    return u.ExprParser(namespace=idents)


def test_parse(parser):
    """Parse."""
    signal = u.Signal(u.UintType(16, default=15), "uint_s")
    expr = parser.parse("uint_s[2]")
    assert expr == u.SliceOp(signal, u.Slice("2"))

    with raises(NameError):
        parser.parse(u.TODO)
    assert parser.parse(2) == u.ConstExpr(u.IntegerType(default=2))
    assert parser.parse(u.IntegerType(default=2)) == u.ConstExpr(u.IntegerType(default=2))

    assert parser.parse("uint_s", only=u.Signal) is signal

    msg = (
        "Signal(UintType(16, default=15), 'uint_s') is not a <class 'ucdp.signal.Port'>. "
        "It is a <class 'ucdp.signal.Signal'>"
    )
    with raises(ValueError, match=re.escape(msg)):
        parser.parse("uint_s", only=u.Port)

    msg = (
        "Signal(UintType(16, default=15), 'uint_s') requires type_ <class 'ucdp.typescalar.SintType'>. "
        "It is a UintType(16, default=15)"
    )
    with raises(ValueError, match=re.escape(msg)):
        parser.parse("uint_s", types=u.SintType)

    msg = "'uint_s *** uint_s': invalid syntax (<string>, line 1)"
    with raises(u.InvalidExpr, match=re.escape(msg)):
        parser.parse("uint_s *** uint_s")


def test_parse_undefined(parser):
    """Undefined names."""
    with raises(NameError, match=re.escape("None: 'uint' is not known.")):
        parser.parse("uint")


# def test_parse_nostrict(idents):
#    """Some Identifier."""
#     parser = u.ExprParser(namespace=idents, strict=False)

#     expr = parser.parse("uint_s[2]")
#     assert expr == u.SliceOp(u.Signal(u.UintType(16, default=15), "uint_s"), u.Slice("2"))

#     assert parser.parse("uint") == "uint"


# def test_parse_empty_nostrict():
#    """Some Identifier."""
#     parser = u.ExprParser(strict=False)

#     expr = parser.parse("uint_s[2]")
#     assert expr == 2
#     assert parser.parse("uint") == "uint"


def test_int():
    """Test Integer."""
    signal = u.Signal(u.UintType(16, default=15), "uint_s")

    expr = signal // 2
    assert expr == u.Op(signal, "//", u.ConstExpr(u.UintType(16, default=2)))

    expr = signal // -2
    assert expr == u.Op(signal, "//", u.ConstExpr(u.SintType(2, default=-2)))


def test_const():
    """Test Constants."""
    assert u.const("0") is u.ConstExpr(u.IntegerType())
    assert u.const("2'b10") is u.ConstExpr(u.UintType(2, default=2))
    assert u.const("-2'sb10") is u.ConstExpr(u.SintType(2, default=-2))

    assert u.const("1'b0") is u.ConstExpr(u.BitType())
    assert u.const("1'b1") is u.ConstExpr(u.BitType(default=1))

    assert u.const("1'h0") is u.ConstExpr(u.UintType(1))
    assert u.const("1'h1") is u.ConstExpr(u.UintType(1, default=1))

    assert u.const(0) is u.ConstExpr(u.IntegerType())
    assert u.const(1) is u.ConstExpr(u.IntegerType(default=1))
    assert u.const(-1) is u.ConstExpr(u.IntegerType(default=-1))

    assert u.const(u.IntegerType.min_) is u.ConstExpr(u.IntegerType(default=u.IntegerType.min_))
    assert u.const(u.IntegerType.max_) is u.ConstExpr(u.IntegerType(default=u.IntegerType.max_))

    assert u.const(u.IntegerType.min_ - 1) is u.ConstExpr(u.SintType(33, default=u.IntegerType.min_ - 1))
    assert u.const(u.IntegerType.max_ + 1) is u.ConstExpr(u.UintType(32, default=u.IntegerType.max_ + 1))

    assert u.const(4) == u.ConstExpr(u.IntegerType(default=4))
    assert u.const("'b10") is u.ConstExpr(u.UintType(2, default=2))
    assert u.const("'b1_0") is u.ConstExpr(u.UintType(2, default=2))
    assert u.const("'o2_1_0") is u.ConstExpr(u.UintType(9, default=136))
    assert u.const("'d1_0") is u.ConstExpr(u.UintType(7, default=10))
    assert u.const("'h4_21_0") is u.ConstExpr(u.UintType(16, default=16912))


def test_concat(parser):
    """Test Concat."""
    expr = parser.parse(("10'd2", 10))
    assert expr == u.ConcatExpr((u.ConstExpr(u.UintType(10, default=2)), u.ConstExpr(u.IntegerType(default=10))))
    assert expr is parser.parse(expr)
    assert expr is parser.concat(expr)
    assert expr[2:1] == u.SliceOp(
        u.ConcatExpr((u.ConstExpr(u.UintType(10, default=2)), u.ConstExpr(u.IntegerType(default=10)))), u.Slice("2:1")
    )
    assert int(expr) == 10242
    assert expr.type_ == u.UintType(
        42,
        default=u.Op(
            u.Op(
                u.Op(
                    u.ConstExpr(u.UintType(10)),
                    "+",
                    u.Op(u.ConstExpr(u.UintType(10, default=2)), "<<", u.ConstExpr(u.UintType(10))),
                ),
                "+",
                u.Op(u.ConstExpr(u.IntegerType(default=10)), "<<", u.ConstExpr(u.IntegerType(default=10))),
            ),
            "+",
            u.ConstExpr(u.UintType(10)),
        ),
    )

    assert u.concat((4, 3, 4)) == u.ConcatExpr(
        (
            u.ConstExpr(u.IntegerType(default=4)),
            u.ConstExpr(u.IntegerType(default=3)),
            u.ConstExpr(u.IntegerType(default=4)),
        )
    )


def test_ternary(parser):
    """Ternary."""
    cond0 = u.Signal(u.BitType(), "if_s") == 0
    cond1 = u.Signal(u.BitType(), "if_s") == 1
    one = u.Signal(u.UintType(16, default=10), "one_s")
    other = u.Signal(u.UintType(16, default=20), "other_s")

    expr = parser.ternary(cond0, one, other)
    assert expr == u.TernaryExpr(
        u.BoolOp(u.Signal(u.BitType(), "if_s"), "==", u.ConstExpr(u.BitType())),
        u.Signal(u.UintType(16, default=10), "one_s"),
        u.Signal(u.UintType(16, default=20), "other_s"),
    )
    assert int(expr) == 10

    expr = parser.ternary(cond1, one, other)
    assert expr == u.TernaryExpr(
        u.BoolOp(u.Signal(u.BitType(), "if_s"), "==", u.ConstExpr(u.BitType(default=1))),
        u.Signal(u.UintType(16, default=10), "one_s"),
        u.Signal(u.UintType(16, default=20), "other_s"),
    )
    assert int(expr) == 20

    a = u.const(4)
    b = u.const(5)
    assert u.ternary(a == b, 4, 5) == u.TernaryExpr(u.BoolOp(a, "==", b), a, b)


def test_note(parser):
    """Note."""
    assert parser.parse_note(u.TODO) is u.TODO
    assert parser.parse_note(2) == u.ConstExpr(u.IntegerType(default=2))
    assert parser.parse_note(u.IntegerType(default=2)) == u.ConstExpr(u.IntegerType(default=2))

    msg = "Default(note='TODO') is not a <class 'ucdp.signal.Signal'>. It is a <class 'ucdp.note.Default'>"
    with raises(ValueError, match=re.escape(msg)):
        parser.parse_note(u.TODO, only=u.Signal)

    msg = "Default(note='TODO') does not meet type_ <class 'ucdp.typescalar.BoolType'>."
    with raises(ValueError, match=re.escape(msg)):
        parser.parse_note(u.TODO, types=u.BoolType)


def test_castbooltype():
    """Cast BoolType."""
    value2 = u.const("2b10")
    boolop = u.BoolOp(value2, "==", value2)
    assert u.cast_booltype(boolop) is boolop

    msg = "ConstExpr(UintType(2, default=2)) does not result in bool"
    with raises(ValueError, match=re.escape(msg)):
        u.cast_booltype(value2)

    boolopuint = u.BoolOp(u.ConstExpr(u.UintType(1, default=1)), "==", u.ConstExpr(u.BitType(default=1)))
    assert u.cast_booltype(u.const("1h1")) == boolopuint

    boolopuint = u.BoolOp(u.ConstExpr(u.UintType(1)), "==", u.ConstExpr(u.UintType(1, default=1)))
    assert u.cast_booltype(u.const("1h0")) == boolopuint


def test_unknown():
    """Unknown."""
    parser = u.ExprParser()
    parser("unknown")


def test_type_error():
    """Type Error."""
    parser = u.ExprParser()
    msg = "'faf' / 5"
    with raises(u.InvalidExpr, match=re.escape(msg)):
        parser("'faf' / 5")


def test_log2():
    """Log 2."""
    a = u.const(67)
    assert u.log2(a) == u.Log2Expr(a)


def test_parse_const(parser):
    """Parse Constants."""
    assert parser("3'd4") == u.ConstExpr(u.UintType(3, default=4))
    assert parser("3'b0") == u.ConstExpr(u.UintType(3, default=0))
    assert parser("'b0") == u.ConstExpr(u.BitType())
    assert parser("'b010") == u.ConstExpr(u.UintType(3, default=2))
    assert parser("'o210") == u.ConstExpr(u.UintType(9, default=136))
    assert parser("'d10") == u.ConstExpr(u.UintType(7, default=10))
    assert parser("'hAF") == u.ConstExpr(u.UintType(8, default=175))



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Module File List."""

from collections.abc import Iterable, Iterator
from inspect import getfile
from pathlib import Path
from typing import Annotated, Any, TypeAlias

from matchor import match
from pydantic import PlainSerializer
from pydantic.functional_validators import BeforeValidator

from .consts import Gen
from .filelistparser import FileListParser
from .iterutil import namefilter
from .modbase import BaseMod, get_modbaseclss
from .moditer import ModPostIter
from .object import Field, IdentObject, Object
from .pathutil import improved_resolve

Paths = tuple[Path, ...]
"""Paths."""

StrPaths = tuple[str | Path, ...]
"""StrPaths."""


def _to_paths(values: Iterable[Any]) -> tuple[Path, ...]:
    if isinstance(values, str):
        return (Path(values),)
    return tuple(Path(value) for value in values)


ToPaths = Annotated[
    StrPaths,
    BeforeValidator(_to_paths),
    PlainSerializer(lambda paths: tuple(path.as_posix() for path in paths), return_type=tuple),
]
"""ToPaths."""

Placeholder = dict[str, Any]
"""
Module Attributes for File Path.

These placeholder are filled during `resolve`.
"""

Flavor: TypeAlias = Object | str
Flavors: TypeAlias = tuple[Flavor, ...]


class ModFileList(IdentObject):
    """
    Module File List.

    Attributes:
        gen: Generate Mode
        targets: Implementation Targets
        inc_dirs: Include Directories
        inc_filepaths: Include File paths relative to module
        filepaths: File paths relative to module
        dep_filepaths: Dependency Filepaths
        dep_inc_dirs: Dependency Include Directories
        template_filepaths: Template Filepaths
        inc_template_filepaths: Template Filepaths
        is_leaf: Do not include file lists of sub modules
    """

    gen: Gen = "no"
    target: str | None = None
    inc_dirs: ToPaths = Field(default=(), strict=False)
    inc_filepaths: ToPaths = Field(default=(), strict=False)
    filepaths: ToPaths = Field(default=(), strict=False)
    dep_inc_dirs: ToPaths = Field(default=(), strict=False)
    dep_filepaths: ToPaths = Field(default=(), strict=False)
    template_filepaths: ToPaths = Field(default=(), strict=False)
    inc_template_filepaths: ToPaths = Field(default=(), strict=False)
    clean_filepaths: ToPaths = Field(default=(), strict=False)
    flavors: Flavors | None = None
    flavor: Flavor | None = None
    is_leaf: bool = False

    @staticmethod
    def get_mod_placeholder(mod: BaseMod, **kwargs) -> Placeholder:
        """Get Module Placeholder."""
        return {"mod": mod, **kwargs}

    @staticmethod
    def get_cls_placeholder(modcls, **kwargs) -> Placeholder:
        """Get Class Placeholder."""
        return {
            "cls": modcls,
            "modref": modcls.get_modref(),
            **kwargs,
        }

    def get_flavors(self, mod: BaseMod) -> Flavors | None:
        """Determine Flavors."""
        return self.flavors

    def generate(self, mod: BaseMod) -> None:
        """Custom Generate Function."""
        raise NotImplementedError

    def get_gen(self, mod: BaseMod, flavor: Flavor | None) -> Gen:
        """Get Generate."""
        return self.gen


ModFileLists = tuple[ModFileList, ...]
"""ModFileLists."""


def search_modfilelists(
    modfilelists: Iterable[ModFileList],
    name: str,
    target: str | None = None,
) -> Iterator[ModFileList]:
    """Search Matching File List.

    Args:
        modfilelists: ModFileLists.
        name: Filelist name or pattern.
        target: Implementation Target

    """
    for modfilelist in modfilelists:
        # Skip Non-Related File Lists
        if not match(modfilelist.name, name):
            continue
        # Skip Non-Matching Target
        if target and modfilelist.target and not namefilter(modfilelist.target)(target):
            continue
        # Found
        yield modfilelist


def resolve_modfilelist(
    mod: BaseMod,
    name: str,
    target: str | None = None,
    filelistparser: FileListParser | None = None,
    replace_envvars: bool = False,
) -> ModFileList | None:
    """Create `ModFileList` for `mod`.

    Args:
        mod: Module.
        name: Name.
        target: Implementation Target
        filelistparser: FileListParser
        replace_envvars: Resolve Environment Variables.
    """
    for modfilelist in resolve_modfilelists(
        mod, name, target=target, filelistparser=filelistparser, replace_envvars=replace_envvars
    ):
        return modfilelist
    return None


def resolve_modfilelists(
    mod: BaseMod,
    name: str,
    target: str | None = None,
    filelistparser: FileListParser | None = None,
    replace_envvars: bool = False,
) -> Iterator[ModFileList]:
    """Create `ModFileList` for `mod`.

    Args:
        mod: Module.
        name: Name or Pattern
        target: Implementation Target
        filelistparser: FileListParser
        replace_envvars: Resolve Environment Variables.
    """
    for modfilelist in search_modfilelists(mod.filelists, name, target=target):
        # parser
        filelistparser = filelistparser or FileListParser()
        for flavor in modfilelist.get_flavors(mod) or [None]:
            # resolve filepaths, inc_dirs
            inc_dirs: list[Path] = []
            inc_filepaths: list[Path] = []
            filepaths: list[Path] = []
            clean_filepaths: list[Path] = []
            mod_placeholder = modfilelist.get_mod_placeholder(mod, flavor=flavor)
            _resolve_mod(
                filelistparser,
                mod,
                mod_placeholder,
                filepaths,
                inc_dirs,
                modfilelist.filepaths,
                modfilelist.inc_dirs,
                replace_envvars,
            )
            _resolve_mod(
                filelistparser,
                mod,
                mod_placeholder,
                inc_filepaths,
                inc_dirs,
                modfilelist.inc_filepaths,
                (),
                replace_envvars,
            )
            # resolve dep_filepaths, dep_inc_dirs
            dep_filepaths: list[Path] = []
            dep_inc_dirs: list[Path] = []
            _resolve_mod(
                filelistparser,
                mod,
                mod_placeholder,
                dep_filepaths,
                dep_inc_dirs,
                modfilelist.dep_filepaths,
                modfilelist.dep_inc_dirs,
                replace_envvars,
            )
            # template_filepaths
            template_filepaths: list[Path] = []
            inc_template_filepaths: list[Path] = []
            baseclss = get_modbaseclss(mod.__class__)
            for basecls in reversed(baseclss):
                for basemodfilelist in search_modfilelists(basecls.filelists, modfilelist.name, target=target):
                    cls_placeholder = basemodfilelist.get_cls_placeholder(basecls, flavor=flavor)
                    _resolve_template_filepaths(
                        basecls,
                        cls_placeholder,
                        template_filepaths,
                        basemodfilelist.template_filepaths,
                        replace_envvars,
                    )
                    _resolve_template_filepaths(
                        basecls,
                        cls_placeholder,
                        inc_template_filepaths,
                        basemodfilelist.inc_template_filepaths,
                        replace_envvars,
                    )
            _resolve_mod(
                filelistparser,
                mod,
                mod_placeholder,
                clean_filepaths,
                (),
                modfilelist.clean_filepaths,
                (),
                replace_envvars,
                glob=True,
            )
            # result
            yield modfilelist.new(
                inc_dirs=tuple(inc_dirs),
                inc_filepaths=tuple(inc_filepaths),
                filepaths=tuple(filepaths),
                dep_filepaths=tuple(dep_filepaths),
                dep_inc_dirs=tuple(dep_inc_dirs),
                template_filepaths=tuple(template_filepaths),
                inc_template_filepaths=tuple(inc_template_filepaths),
                clean_filepaths=tuple(clean_filepaths),
                flavors=(flavor,) if flavor is not None else None,
                flavor=flavor,
            )


def iter_modfilelists(
    topmod: BaseMod,
    name: str,
    target: str | None = None,
    filelistparser: FileListParser | None = None,
    replace_envvars: bool = False,
    maxlevel: int | None = None,
) -> Iterator[tuple[BaseMod, ModFileList]]:
    """Iterate over `ModFileLists`.

    Args:
        topmod: Top Module.
        name: Name or Name Pattern.
        target: Implementation Target
        filelistparser: FileListParser
        replace_envvars: Resolve Environment Variables.
        maxlevel: Stop at maximum iteration level.
    """
    filelistparser = filelistparser or FileListParser()

    # stop at leaf
    def stop_insts(inst: BaseMod):
        for modfilelist in search_modfilelists(inst.filelists, name, target=target):
            if modfilelist.is_leaf:
                return True
        return False

    # iterate
    for mod in ModPostIter(topmod, stop_insts=stop_insts, unique=True, maxlevel=maxlevel):
        for modfilelist in resolve_modfilelists(
            mod,
            name=name,
            target=target,
            filelistparser=filelistparser,
            replace_envvars=replace_envvars,
        ):
            yield mod, modfilelist


def _resolve_mod(
    filelistparser: FileListParser,
    mod: BaseMod,
    placeholder: Placeholder,
    filepaths: list[Path],
    inc_dirs: list[Path],
    add_filepaths: StrPaths,
    add_inc_dirs: StrPaths,
    replace_envvars: bool,
    glob: bool = False,
) -> None:
    basefile = Path(getfile(mod.__class__))
    basedir = basefile.parent
    if add_inc_dirs:
        items = (Path(str(filepath).format_map(placeholder)) for filepath in add_inc_dirs)
        filelistparser.parse(
            inc_dirs, inc_dirs, basedir, items, replace_envvars=replace_envvars, context=str(basefile), glob=glob
        )
    if add_filepaths:
        items = (Path(str(filepath).format_map(placeholder)) for filepath in add_filepaths)
        filelistparser.parse(
            filepaths,
            inc_dirs,
            basedir,
            items,
            replace_envvars=replace_envvars,
            context=str(basefile),
            glob=glob,
        )


def _resolve_template_filepaths(
    cls,  # class BaseMod
    placeholder: Placeholder,
    filepaths: list[Path],
    add_filepaths: StrPaths,
    replace_envvars: bool,
):
    basedir = Path(getfile(cls)).parent
    if add_filepaths:
        items = tuple(Path(str(item).format_map(placeholder)) for item in add_filepaths)
        for add_filepath in reversed(items):
            try:
                filepath = improved_resolve(
                    add_filepath,
                    basedir=basedir,
                    replace_envvars=replace_envvars,
                    strict=replace_envvars,
                )
            except FileNotFoundError:
                # Template is found through search path
                filepath = add_filepath
            if filepath not in filepaths:
                filepaths.insert(0, filepath)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Type Utilities.

>>> import ucdp as u
>>> from tabulate import tabulate
>>> class UEnumType(u.AEnumType):
...     keytype: u.BaseScalarType = u.UintType(2)
...     def _build(self):
...         self._add(0, 'one')
...         self._add(1, 'two')
>>> class SEnumType(u.AEnumType):
...     keytype: u.BaseScalarType = u.SintType(2)
...     def _build(self):
...         self._add(0, 'one')
...         self._add(-1, 'two')
>>> types = (
...     u.BitType(),
...     u.UintType(10),
...     u.SintType(10),
...     u.BoolType(),
...     u.RailType(),
...     UEnumType(),
...     SEnumType(),
...     u.ClkRstAnType(),
...     u.ArrayType(u.UintType(10),5),
...     u.ArrayType(u.SintType(10),5),
... )

>>> print(tabulate([(t, u.is_scalar(t), u.is_signed(t)) for t in types], headers=("type", "is_scalar", "is_signed")))
type                        is_scalar    is_signed
--------------------------  -----------  -----------
BitType()                   True         False
UintType(10)                True         False
SintType(10)                True         True
BoolType()                  True         False
RailType()                  True         False
UEnumType()                 True         False
SEnumType()                 True         True
ClkRstAnType()              False
ArrayType(UintType(10), 5)  True         False
ArrayType(SintType(10), 5)  True         True
"""

from .typearray import ArrayType
from .typebase import AScalarType, BaseType
from .typebaseenum import BaseEnumType
from .typescalar import IntegerType, SintType


def is_scalar(type_: BaseType) -> bool:
    """Check if `type` is scalar."""
    if isinstance(type_, ArrayType):
        return is_scalar(type_.itemtype)
    if isinstance(type_, BaseEnumType):
        return is_scalar(type_.keytype)
    return isinstance(type_, AScalarType)


def is_signed(type_: BaseType) -> bool | None:
    """
    Check if `type` is signed.
    """
    if isinstance(type_, AScalarType):
        return isinstance(type_, (SintType, IntegerType))
    if isinstance(type_, ArrayType):
        return is_signed(type_.itemtype)
    if isinstance(type_, BaseEnumType):
        return is_signed(type_.keytype)
    return None



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Base Top Module.
"""

from abc import abstractmethod
from typing import Any, ClassVar

from ._modbuilder import build
from .modbase import BaseMod
from .modfilelist import ModFileLists
from .modutil import get_libpath, is_tb_from_modname
from .object import PrivateField


class BaseTopMod(BaseMod):
    """Base Class For All Top Modules."""

    filelists: ClassVar[ModFileLists] = ()
    """File Lists."""

    _has_build_final: bool = PrivateField(default=True)

    @abstractmethod
    def _build(self) -> None:
        """Build Type."""

    def _build_final(self) -> None:
        """Build Post."""

    def model_post_init(self, __context: Any) -> None:
        """Run Build."""
        if self.parent:
            self.parent.add_inst(self)

        self._build()
        build(self)

    @property
    def libpath(self) -> str:
        """Library Path."""
        return get_libpath(self.__class__)

    @property
    def is_tb(self) -> bool:
        """Determine if module belongs to Testbench or Design."""
        return is_tb_from_modname(self.modname)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Struct Type Testing.
"""

import re

from pytest import raises

import ucdp as u


def test_struct():
    """Struct."""

    class MyStructType(u.AStructType):
        """Struct."""

        def _build(self) -> None:
            self._add("data", u.UintType(8), title="title")
            self._add("valid", u.BitType(), descr="descr", ifdefs=("IFDEF",))
            self._add("accept", u.BitType(), u.BWD, comment="comment")

            msg = (
                "name 'valid' already exists in test_struct.<locals>.MyStructType() "
                "(StructItem('valid', BitType(), doc=Doc(descr='descr'), ifdefs=('IFDEF',)))"
            )
            with raises(ValueError, match=re.escape(msg)):
                self._add("valid", u.BitType())

    struct = MyStructType()

    msg = "test_struct.<locals>.MyStructType(): Cannot add item 'lock'."
    with raises(u.LockError, match=re.escape(msg)):
        struct._add("lock", u.BitType())

    assert tuple(struct) == tuple(struct.keys())
    assert tuple(struct.keys()) == ("data", "valid", "accept")
    assert tuple(struct.values()) == (
        u.StructItem("data", u.UintType(8), doc=u.Doc(title="title")),
        u.StructItem("valid", u.BitType(), doc=u.Doc(descr="descr"), ifdefs=("IFDEF",)),
        u.StructItem("accept", u.BitType(), orientation=u.BWD, doc=u.Doc(comment="comment")),
    )
    assert [item.doc for item in struct.values()] == [
        u.Doc(title="title"),
        u.Doc(descr="descr"),
        u.Doc(comment="comment"),
    ]

    assert MyStructType() is MyStructType()
    assert MyStructType() == MyStructType()


def test_inherit_basestruct():
    """:any:`BaseStructType` must be inherited."""
    with raises(TypeError):
        u.BaseStructType()


def test_inherit_struct():
    """:any:`StructType` must be inherited."""
    with raises(TypeError):
        u.AStructType()


def test_inherit_globalstruct():
    """:any:`AGlobalStructType` must be inherited."""
    with raises(TypeError):
        u.AGlobalStructType()


def test_inherit_dynamicstruct():
    """:any:`DynamicStructType` must be inherited."""
    u.DynamicStructType()


def test_globalstruct():
    """Global Struct."""

    class MyType(u.AGlobalStructType):
        pass

    one = MyType()
    one.add("data", u.UintType(8))
    one.add("valid", u.BitType())
    assert tuple(one) == ("data", "valid")

    other = MyType()
    other.add("accept", u.BitType(), orientation=u.BWD)

    assert one is other
    assert one == other
    assert tuple(one) == ("data", "valid", "accept")
    assert tuple(other) == ("data", "valid", "accept")


def test_dynamicstruct():
    """Dynamic Struct."""

    class MyType(u.DynamicStructType):
        pass

    one = MyType()
    one.add("data", u.UintType(8))
    one.add("valid", u.BitType())
    assert tuple(one) == ("data", "valid")

    other = MyType()
    other.add("accept", u.BitType(), orientation=u.BWD)

    assert one is not other
    assert one != other
    assert tuple(one) == ("data", "valid")
    assert tuple(other) == ("accept",)


def test_structitem():
    """StructItem testing."""
    doc = u.Doc(title="title", descr="descr", comment="comment")
    item = u.StructItem(
        "data",
        u.UintType(8),
        doc=doc,
    )

    assert item.doc is doc
    assert item.title == "title"
    assert item.descr == "descr"
    assert item.comment == "comment"


class ClkRelStruct(u.AStructType):
    """Example Clock Relation."""

    def _build(self):
        self._add("clk0", u.ClkType())
        self._add("data0", u.BitType(), clkrel=u.ClkRel(clk="clk0"))
        self._add("clk1", u.ClkType())
        self._add("data1", u.BitType(), clkrel=u.ClkRel(clk="clk1"), orientation=u.BWD)


def test_clkrel_ident():
    """Clock Relation."""
    ident = u.Ident(ClkRelStruct(), "name")
    assert tuple(ident.iter()) == (
        u.Ident(ClkRelStruct(), "name"),
        u.Ident(u.ClkType(), "name_clk0", doc=u.Doc(title="Clock")),
        u.Ident(u.BitType(), "name_data0"),
        u.Ident(u.ClkType(), "name_clk1", doc=u.Doc(title="Clock")),
        u.Ident(u.BitType(), "name_data1"),
    )


def test_clkrel_port():
    """Clock Relation."""
    port = u.Port(ClkRelStruct(), "port_o")
    assert tuple(port.iter()) == (
        u.Port(ClkRelStruct(), "port_o", direction=u.OUT),
        u.Port(u.ClkType(), "port_clk0_o", direction=u.OUT, doc=u.Doc(title="Clock")),
        u.Port(u.BitType(), "port_data0_o", direction=u.OUT, clkrel=u.ClkRel(clk="clk0")),
        u.Port(u.ClkType(), "port_clk1_o", direction=u.OUT, doc=u.Doc(title="Clock")),
        u.Port(u.BitType(), "port_data1_i", direction=u.IN, clkrel=u.ClkRel(clk="clk1")),
    )



"""Mod5 Module."""


from fileliststandard import HdlFileList
from glbl_lib.bus import BusType
from glbl_lib.clk_gate import ClkGateMod
from glbl_lib.regf import RegfMod

import ucdp as u


class Mod5Mod(u.ATbMod):
    """Mod5 Module."""

    filelists: u.ClassVar[u.ModFileLists] = (
        HdlFileList(gen="full"),
    )

    def _build(self) -> None:
        dut = self.dut # Design-Under-Test



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Name Utilities.
"""

import functools
import re
from collections.abc import Iterable
from typing import Any

from caseconverter import snakecase
from fuzzywuzzy import process
from humanfriendly.text import concatenate

_FUZZY_MINRATIO: int = 80
_RE_STARTNUM = re.compile(r"^[0-9]")
_RE_SPLIT_PREFIX = re.compile(r"(?i)(?P<prefix>([a-z])_)(?P<basename>([a-z][a-z0-9_]*)?)\Z")
_RE_SPLIT_SUFFIX = re.compile(r"(?i)(?P<basename>([a-z][a-z0-9_]*)?)(?P<suffix>_([a-z][o]?))\Z")


@functools.lru_cache
def split_prefix(name: str) -> tuple[str, str]:
    """
    Split Name Into Prefix and Basename.

    Args:
        name: Name.

    Returns:
        tuple: Tuple of Prefix and Basename

    ??? Example "split_prefix Examples"
        Example:

            >>> split_prefix("i_count")
            ('i_', 'count')
            >>> split_prefix("u_count")
            ('u_', 'count')
            >>> split_prefix("I_VERY_LONG_NAME")
            ('I_', 'VERY_LONG_NAME')
            >>> split_prefix("")
            ('', '')

            The counterpart to this function is `join_names`.
    """
    mat = _RE_SPLIT_PREFIX.match(name)
    if mat:
        return mat.group("prefix", "basename")  # type: ignore[return-value]
    return "", name


@functools.lru_cache
def split_suffix(name: str, only: tuple[str, ...] = ()) -> tuple[str, str]:
    """
    Split Name Into Basename And Suffix.

    Args:
        name: Name.

    Keyword Args:
        only: Limit suffix to given.

    Returns:
        tuple: Tuple of Basename and Suffix

    ??? Example "split_suffix Examples"
        Example:

            >>> split_suffix("count_i")
            ('count', '_i')
            >>> split_suffix("count_o")
            ('count', '_o')
            >>> split_suffix("count_io")
            ('count', '_io')
            >>> split_suffix("count_t")
            ('count', '_t')
            >>> split_suffix("count_s")
            ('count', '_s')
            >>> split_suffix("_s")
            ('', '_s')
            >>> split_suffix("very_long_name_s")
            ('very_long_name', '_s')
            >>> split_suffix("VERY_LONG_NAME_S")
            ('VERY_LONG_NAME', '_S')
            >>> split_suffix("")
            ('', '')

            The counterpart to this function is `join_names`.
    """
    mat = _RE_SPLIT_SUFFIX.match(name)
    if mat:
        basename, suffix = mat.group("basename", "suffix")  # type: ignore[return-value]
        if not only or suffix in only:
            return basename, suffix
    return name, ""


def join_names(*names: str, concat: str = "_") -> str:
    """
    Join Names.

    Args:
        names: Names.

    Keyword Args:
        concat: concat.

    ??? Example "join_names Examples"
        Example:

            >>> join_names('foo', 'bar')
            'foo_bar'
            >>> join_names('', 'bar')
            'bar'
            >>> join_names('foo', '')
            'foo'

    This function is the counterpart to `split_names`.
    """
    return concat.join(name for name in names if name)


def didyoumean(name: str, names: Iterable[str], known: bool = False, multiline: bool = False) -> str:
    """
    Propose matching names.

    Args:
        name: Name.
        names: Names.

    Keyword Args:
        known: Show known strings.
        multiline: Add new line

    ??? Example "didyoumean Examples"
        Example:

            >>> didyoumean('abb', tuple())
            ''
            >>> didyoumean('abb', ('abba', 'ac/dc', 'beatles'), known=True)
            " Known are 'abba', 'ac/dc' and 'beatles'. Did you mean 'abba'?"
            >>> print(didyoumean('zz-top', ('abba', 'ac/dc', 'beatles'), known=True, multiline=True))
            <BLANKLINE>
            Known are 'abba', 'ac/dc' and 'beatles'.
    """
    sep = "\n" if multiline else " "
    if known:
        knowns = concatenate(repr(name) for name in names)
        msg = f"{sep}Known are {knowns}."
    else:
        msg = ""
    # fuzzy
    if names:
        fuzzyitems = (repr(item) for item, ratio in process.extract(name, names, limit=5) if ratio >= _FUZZY_MINRATIO)
        fuzzy = concatenate(fuzzyitems, conjunction="or")
        if fuzzy:
            msg = f"{msg}{sep}Did you mean {fuzzy}?"
    return msg


def get_snakecasename(item: Any):
    """
    Get snakecase name of `cls`.

    ??? Example "get_snakecasename Examples"
        Example:

            >>> class MyClass:
            ...     pass
            >>> get_snakecasename(MyClass)
            'my_class'
            >>> get_snakecasename("MyClass")
            'my_class'
            >>> get_snakecasename("MYClass")
            'myclass'
    """
    if not isinstance(item, str):
        item = item.__name__
    return snakecase(item).removeprefix("_")


def str2identifier(value: str) -> str:
    """
    Convert Any String To Identifier.

    Args:
        value: Any String

    ??? Example "str2identifier Examples"
        Example:

            >>> str2identifier('A B C')
            'a_b_c'
            >>> str2identifier('1 2 3')
            'v1_2_3'
            >>> str2identifier('12.3')
            'v123'
    """
    value = snakecase(value)
    mat = _RE_STARTNUM.match(value)
    if mat:
        value = f"v{value}"
    return value



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Module File Information."""

import re

from pytest import raises
from test2ref import assert_refdata

import ucdp as u


class FlipFlopMod(u.AMod):
    """Module using FlipFlop."""

    def _build(self):
        self.add_port(u.ClkRstAnType(), "main_i")
        self.add_port(u.UintType(2), "ctrl_i")
        self.add_port(u.UintType(8), "data_i")
        self.add_port(u.UintType(8), "data_o")

        self.add_flipflop(u.UintType(8), "data_r", "main_clk_i", "main_rst_an_i", nxt="data_i")
        self.add_flipflop(u.UintType(8), "other_r", "main_clk_i", "main_rst_an_i", ena="ctrl_i == 1", rst="ctrl_i == 0")

        msg = "clk is not of ClkType"
        with raises(ValueError, match=re.escape(msg)):
            self.add_flipflop(u.UintType(8), "noclk_r", "ctrl_i", "main_rst_an_i")

        msg = "rst_an is not of RstAnType"
        with raises(ValueError, match=re.escape(msg)):
            self.add_flipflop(u.UintType(8), "norst_r", "main_clk_i", "ctrl_i")


def test_flipflop(tmp_path, capsys):
    """FlipFlop."""
    mod = FlipFlopMod()

    _print_flipflops(mod)
    assert_refdata(test_flipflop, tmp_path, capsys=capsys)


def _print_flipflops(mod: u.BaseMod):
    rslvr = u.ExprResolver()
    for flipflop in mod.flipflops:
        print("clk", rslvr(flipflop.clk))
        print("rst_an", rslvr(flipflop.rst_an))
        print("rst", rslvr(flipflop.rst) if flipflop.rst else "-")
        print("ena", rslvr(flipflop.ena) if flipflop.ena else "-")
        for item in flipflop:
            print(f"  {item}")



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Testbench Module.
"""

from .config import BaseConfig
from .modbase import BaseMod
from .modtb import ATbMod
from .modutil import get_modname, get_topmodname
from .nameutil import join_names


class AConfigurableTbMod(ATbMod):
    """
    A Testbench Module Which Is Assembled According To A Recipe ([AConfig][ucdp.config.AConfig]).
    """

    config: BaseConfig
    is_default: bool

    def __init__(self, parent: BaseMod | None = None, name: str | None = None, **kwargs):
        is_default = False
        if "config" not in kwargs:
            try:
                kwargs["config"] = self.get_default_config()
                is_default = True
            except NotImplementedError:
                pass
        try:
            is_default = is_default or kwargs["config"].is_default
        except KeyError:
            if parent is not None:
                raise ValueError("'config' is required if 'parent' is given") from None
        super().__init__(parent=parent, name=name, is_default=is_default, **kwargs)  # type: ignore[call-arg]

    @property
    def modname(self) -> str:
        """Module Name."""
        config = self.config
        if config.modname:
            return config.modname
        name = config.name
        modbasename = get_modname(self.__class__)
        if not name and "config" in self.model_fields_set and not self.is_default:
            name = config.unique_name
        return join_names(modbasename, name)

    @property
    def topmodname(self) -> str:
        """Top Module Name."""
        return get_topmodname(self.__class__)

    def get_default_config(self) -> BaseConfig:
        """Create Default Configuration."""
        raise NotImplementedError



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Loading And Searching Facility.
"""

import re
import sys
from collections.abc import Iterable, Iterator
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache
from importlib import import_module
from inspect import getfile, isclass
from pathlib import Path
from typing import TypeAlias

from .consts import PKG_PATHS
from .modbase import BaseMod, get_modbaseclss
from .modref import ModRef, get_modclsname
from .modtopref import TopModRef
from .object import Object
from .pathutil import absolute
from .util import LOGGER, get_maxworkers, guess_path

Patterns: TypeAlias = Iterable[str]
Paths: TypeAlias = Iterable[Path]

_RE_IMPORT_UCDP = re.compile(r"^\s*class .*Mod\):")

_RE_TOPMODREFPAT = re.compile(
    # [tb]#
    r"((?P<tb>[a-zA-Z_0-9_\.\*]+)#)?"
    # top
    r"(?P<top>[a-zA-Z_0-9_\.\*]+)"
    # [-sub]
    r"(-(?P<sub>[a-zA-Z_0-9_\.\*]+))?"
)


class TopModRefPat(Object):
    """Top Module Reference Search Pattern Pattern."""

    top: str
    sub: str | None = None
    tb: str | None = None

    def __str__(self):
        result = self.top
        if self.sub:
            result = f"{result}-{self.sub}"
        if self.tb:
            result = f"{self.tb}#{result}"
        return result


@lru_cache
def build_top(modcls, **kwargs):
    """Build Top Module."""
    return modcls.build_top(**kwargs)


@lru_cache
def load_modcls(modref: ModRef) -> type[BaseMod]:
    """Load Module Class."""
    name = f"{modref.libname}.{modref.modname}"
    try:
        pymod = import_module(name)
    except ModuleNotFoundError as exc:
        if exc.name in (modref.libname, name):
            raise NameError(f"{name!r} not found.") from None
        raise exc
    modclsname = modref.get_modclsname()
    modcls = getattr(pymod, modclsname, None)
    if not modcls:
        raise NameError(f"{name!r} does not contain {modclsname}.") from None
    if not issubclass(modcls, BaseMod):
        raise ValueError(f"{modcls} is not a module aka child of <class ucdp.BaseMod>.")
    return modcls


def find_modrefs(local: bool | None = None) -> tuple[ModRef, ...]:
    # determine directories with python files
    dirpaths: set[Path] = set()
    modrefs: list[ModRef] = []
    for syspathstr in sys.path:
        syspath = absolute(Path(syspathstr))
        if local is not None and local is any(syspath.is_relative_to(pkg_path) for pkg_path in PKG_PATHS):
            continue
        for filepath in syspath.glob("*/*.py"):
            dirpath = filepath.parent
            if dirpath.name.startswith("_") or dirpath.name == "ucdp":
                continue
            dirpaths.add(dirpath)

    maxworkers = get_maxworkers()
    with ThreadPoolExecutor(max_workers=maxworkers) as exe:
        # start
        jobs = [
            exe.submit(_find_modrefs, sys.path, tuple(sorted(dirpath.glob("*.py")))) for dirpath in sorted(dirpaths)
        ]
        # collect
        for job in jobs:
            modrefs.extend(job.result())
    return tuple(modrefs)


def _find_modrefs_files(modrefs: tuple[ModRef, ...], envpath: list[str], filepaths: tuple[Path, ...]) -> list[Path]:
    paths: set[Path] = set(filepaths)
    for modref in modrefs:
        modcls = load_modcls(modref)
        for basecls in get_modbaseclss(modcls):
            paths.add(Path(getfile(basecls)))
    return sorted(paths)


def _find_modrefs(envpath: list[str], filepaths: tuple[Path, ...]) -> tuple[ModRef, ...]:  # noqa: C901
    modrefs = []
    for filepath in filepaths:
        pylibname = filepath.parent.name
        pymodname = filepath.stem
        if pymodname.startswith("_"):
            continue
        # skip non-ucdp files
        try:
            with filepath.open(encoding="utf-8") as file:
                for line in file:
                    if _RE_IMPORT_UCDP.match(line):
                        break
                else:
                    continue
        except Exception as exc:
            LOGGER.info(f"Skipping {str(filepath)!r} ({exc})")
            continue

        # import module
        try:
            pymod = import_module(f"{pylibname}.{pymodname}")
        except Exception as exc:
            LOGGER.warning(f"Skipping {str(filepath)!r} ({exc})")
            continue

        # Inspect Module
        for name in dir(pymod):
            # Load Class
            modcls = getattr(pymod, name)
            if not isclass(modcls) or not issubclass(modcls, BaseMod):
                continue

            # Ignore imported
            if filepath != Path(getfile(modcls)):
                continue

            # Create ModRefInfo
            modclsname = get_modclsname(pymodname)
            if modclsname == name:
                modref = ModRef(libname=pylibname, modname=pymodname)
            else:
                modref = ModRef(libname=pylibname, modname=pymodname, modclsname=name)
            modrefs.append(modref)

    return tuple(modrefs)


def get_topmodrefpats(patterns: Patterns | None) -> Iterator[TopModRefPat | TopModRef]:
    for pattern in patterns or []:
        path = guess_path(pattern)
        if path:
            yield TopModRef.cast(path)
        else:
            mat = _RE_TOPMODREFPAT.fullmatch(pattern)
            if mat:
                yield TopModRefPat(**mat.groupdict())
            else:
                yield TopModRefPat(top=".")  # never matching



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
UCDP Base Objects.

There are two base objects:

* [Object][ucdp.object.Object]
* [LightObject][ucdp.object.LightObject]

Every UCDP object must be derived from [Object][ucdp.object.Object].

DOCME: what for what
DOCME: pydantic, strict, no extra
DOCME: use examples
DOCME: caching
"""

from typing import Any, ClassVar, TypeAlias

import pydantic as pyd
from pydantic._internal._model_construction import ModelMetaclass

from .consts import PAT_IDENTIFIER

_cache: dict[tuple[Any, ...], Any] = {}

PosArgs: TypeAlias = tuple[str, ...]

Field = pyd.Field
PrivateField = pyd.PrivateAttr
ConfigDict = pyd.ConfigDict
model_validator = pyd.model_validator
computed_field = pyd.computed_field

_CACHED_INSTANCES: int = 0


class Object(pyd.BaseModel):
    """Read-Only :any:`pydantic` Base Model."""

    model_config = pyd.ConfigDict(
        extra="forbid",
        frozen=True,
        revalidate_instances="never",
        strict=True,
        validate_default=True,
        arbitrary_types_allowed=True,
    )

    _posargs: ClassVar[PosArgs] = ()

    def new(self, **kwargs):
        """Return A Copy With Updated Attributes."""
        data = {}
        values = self.__dict__
        for name in self.model_fields_set:
            data[name] = values[name]
        data.update(kwargs)
        return self.__class__(**data)

    def __str__(self) -> str:
        """String."""
        return get_repr(self)

    def __repr__(self) -> str:
        """Representation."""
        return get_repr(self)


class CachedModelMetaclass(ModelMetaclass):
    """Meta Class for Cached Model Instances."""

    _cache: ClassVar[dict[tuple[Any, ...], Any]] = {}

    def __call__(self, *args, **kwargs):
        """Create New Instance or Return Existing One."""
        key = (self, *args, *sorted(kwargs.items()))
        try:
            try:
                inst = self._cache[key]
            except KeyError:
                global _CACHED_INSTANCES  # noqa: PLW0603
                _CACHED_INSTANCES += 1
                inst = self._cache[key] = super().__call__(*args, **kwargs)
        except TypeError as exc:
            try:
                hash(self)
            except TypeError:  # pragma: no cover
                raise TypeError(f"{self} is not constant.") from None
            # Determine what caused TypeError
            for idx, arg in enumerate(args):
                try:
                    hash(arg)
                except TypeError:  # noqa: PERF203
                    raise TypeError(f"{self}: {idx} argument {arg!r} is not constant.") from None
            for name, value in kwargs.items():
                try:
                    hash(name)
                    hash(value)
                except TypeError:  # noqa: PERF203
                    raise TypeError(f"{self}: {name!r} argument {value!r} is not constant.") from None
            raise exc
        return inst


class Light(metaclass=CachedModelMetaclass):
    """DOCME."""


class LightObject(Object, Light):
    """DOCME."""


class NamedObject(Object):
    """NamedObject.

    Attributes:
        name: Name.
    """

    name: str


class NamedLightObject(NamedObject, Light):
    """Cacheable NamedObject.

    Attributes:
        name: Name.
    """


class IdentObject(NamedObject):
    """
    Identifier Object.

    Attributes:
        name: Name.
    """

    name: str = Field(pattern=PAT_IDENTIFIER)


class IdentLightObject(IdentObject, Light):
    """Cacheable IdentObject.

    Attributes:
        name: Name.
    """


def get_repr(obj) -> str:
    """DOCME."""
    posargs = obj._posargs
    model_fields_set = obj.model_fields_set
    values = obj.__dict__
    sign_args = tuple(repr(values[key]) for key in posargs)
    sign_kwargs = (
        f"{key}={values[key]!r}"
        for key, field in obj.__class__.model_fields.items()
        if field.repr and key in model_fields_set and key not in posargs and values[key] != field.default
    )
    sign = ", ".join((*sign_args, *sign_kwargs))
    return f"{obj.__class__.__qualname__}({sign})"



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Path Utilities."""

import os
import os.path
import re
from collections.abc import Iterator
from pathlib import Path

_RE_PAT = re.compile(r".*[\*\?\]\[]")
_RE_ENVVAR = re.compile(
    r"^\$(?P<name>(\{(?P<name1>[^\/\\]+)\})|(?P<name0>[^\/\\]+))([\/\\](?P<rem>.*))?$", flags=re.IGNORECASE
)


def improved_glob(pattern: Path | str, basedir: Path | None = None) -> Iterator[Path]:
    """
    Improved version of pathlib.Path.glob().

    * Not existing files are not 'globbed-away'
    * Aware of Environment Variables
    * Output is sorted.
    """
    basedir = basedir or Path()
    patternstr = str(pattern)
    if _is_pattern(patternstr):
        name, path = startswith_envvar(Path(pattern))
        if name:
            vname = f"${name}"
            base = Path(vname)
            value = Path(os.path.expandvars(vname))
            for sub in sorted(value.glob(str(path))):
                rel = sub.relative_to(value)
                yield base / rel
        else:
            yield from sorted(basedir.glob(patternstr))
    else:
        yield basedir / pattern


def improved_resolve(path: Path, basedir: Path | None = None, strict: bool = False, replace_envvars=False) -> Path:
    """
    Improved version of pathlib.Path.resolve().

    * Aware of Environment Variables
    """
    if replace_envvars:
        path = Path(os.path.expandvars(str(path)))
    elif _RE_ENVVAR.match(str(path)):
        if strict:
            raise FileNotFoundError(path)
        return path

    if not path.is_absolute():
        basedir = basedir or Path()
        path = basedir / path
    path = absolute(path)
    if strict:
        path.resolve(strict=True)
    return path


def use_envvars(path: Path, envvarnames: tuple[str, ...]) -> Path:
    """
    Use Environment Variables instead of absolute Paths.

    Do nothing on relative paths.
    """
    if not path.is_absolute():
        return path
    for envvarname in envvarnames:
        value = os.getenv(envvarname)
        if not value:
            continue
        try:
            sub = path.relative_to(Path(value))
        except ValueError:
            continue
        return Path(f"${{{envvarname}}}") / sub
    return path


def startswith_envvar(path: Path, strict: bool = False, barename: bool = False) -> tuple[str | None, Path]:
    """Check if path starts with environment variable."""
    mat = _RE_ENVVAR.match(str(path))
    if not mat:
        return None, path
    varname = mat.group("name0") or mat.group("name1")
    path = Path(mat.group("rem") or ".")
    if strict:
        envpath = os.getenv(varname)
        if not envpath or not Path(envpath).exists():
            raise FileNotFoundError(envpath)

    if barename:
        return varname, path
    return mat.group("name"), path


def _is_pattern(pattern: str) -> bool:
    return bool(_RE_PAT.match(pattern))


def relative(path: Path, base: Path | None = None) -> Path:
    """Relative."""
    base = base or absolute(Path())
    try:
        return path.relative_to(base)
    except ValueError:
        return Path(os.path.relpath(str(path), str(base)))


def absolute(path: Path) -> Path:
    """Resolve `..`, but keep symlinks."""
    return Path(os.path.abspath(path))  # noqa: PTH100



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Route Path."""

import re

from pytest import raises

import ucdp as u


def test_parsepath_full():
    """Full Path."""
    path = u.parse_routepath("u_sub/u_bar/clk_i")
    assert str(path) == "u_sub/u_bar/clk_i"
    assert path.path == "u_sub/u_bar"
    assert path.expr == "clk_i"
    assert path.create is False
    assert path.cast is False


def test_parsepath_empty():
    """Empty Path."""
    with raises(ValueError, match=re.escape("Invalid route path ''")):
        u.parse_routepath("")


def test_parsepath_up():
    """Up Path."""
    path = u.parse_routepath("../port_i")
    assert str(path) == "../port_i"
    assert path.path == ".."
    assert path.expr == "port_i"
    assert path.create is False
    assert path.cast is False


def test_parsepath_cast():
    """Casting."""
    path = u.parse_routepath("cast(u_sub/clk_i)")
    assert str(path) == "cast(u_sub/clk_i)"
    assert path.path == "u_sub"
    assert path.expr == "clk_i"
    assert path.create is False
    assert path.cast is True


def test_parsepath_optcast():
    """'Optional Casting."""
    path = u.parse_routepath("optcast(u_sub/clk_i)")
    assert str(path) == "optcast(u_sub/clk_i)"
    assert path.path == "u_sub"
    assert path.expr == "clk_i"
    assert path.create is False
    assert path.cast is None


def test_parsepath_create():
    """Creating."""
    path = u.parse_routepath("create(u_sub/clk_i)")
    assert str(path) == "create(u_sub/clk_i)"
    assert path.path == "u_sub"
    assert path.expr == "clk_i"
    assert path.create is True
    assert path.cast is False


def test_parsepath_create_cast():
    """Create and Casting."""
    with raises(ValueError, match=re.escape("Invalid route path 'create(cast((u_sub/clk_i))'")):
        u.parse_routepath("create(cast((u_sub/clk_i))")


def test_parsepath_create_cast_bracket():
    """Create and Casting."""
    with raises(ValueError, match=re.escape("Invalid route path 'cast(u_sub/clk_i'")):
        u.parse_routepath("cast(u_sub/clk_i")


def test_parsepath_basepath():
    """Basepath Path."""
    path = u.parse_routepath("u_sub/u_bar/clk_i", basepath="u_sim")
    assert str(path) == "u_sim/u_sub/u_bar/clk_i"
    assert path.path == "u_sim/u_sub/u_bar"
    assert path.expr == "clk_i"
    assert path.create is False
    assert path.cast is False


def test_create_cast():
    """Create and Casting."""
    with raises(u.ValidationError, match=re.escape("[opt]cast() and create() are mutually exclusive")):
        u.RoutePath(expr="clk_i", create=True, cast=True)


def test_expr():
    """Expression."""
    expr = u.Signal(u.UintType(8), "sig_s")
    path = u.parse_routepath(expr)
    assert str(path) == "sig_s"
    assert path.path is None
    assert path.expr is expr
    assert path.create is False
    assert path.cast is False
